"""
Abstract base class for the pointwise safeguards.
"""

__all__ = ["PointwiseSafeguard"]

from abc import ABC, abstractmethod
from typing import Literal, final

import numpy as np
from typing_extensions import override  # MSPV 3.12

from ...utils.bindings import Bindings
from ...utils.intervals import IntervalUnion
from ...utils.typing import S, T
from ..abc import Safeguard


class PointwiseSafeguard(Safeguard, ABC):
    """
    Pointwise safeguard abstract base class.

    Pointwise safeguards describe properties that are satisfied (or not) per
    element, i.e. independent of other elements.
    """

    __slots__: tuple[str, ...] = ()

    @final
    @override
    def check(
        self,
        data: np.ndarray[S, np.dtype[T]],
        prediction: np.ndarray[S, np.dtype[T]],
        *,
        late_bound: Bindings,
        where: Literal[True] | np.ndarray[S, np.dtype[np.bool]] = True,
    ) -> bool:
        """
        Check if the `prediction` array upholds the property enforced by this
        safeguard.

        Parameters
        ----------
        data : np.ndarray[S, np.dtype[T]]
            Original data array, relative to which the `prediction` is checked.
        prediction : np.ndarray[S, np.dtype[T]]
            Prediction for the `data` array.
        late_bound : Bindings
            Bindings for late-bound parameters, including for this safeguard.
        where : Literal[True] | np.ndarray[S, np.dtype[np.bool]]
            Only check at data points where the condition is [`True`][True].

        Returns
        -------
        ok : bool
            `True` if the check succeeded.
        """

        return bool(
            np.all(self.check_pointwise(data, prediction, late_bound=late_bound))
        )

    @abstractmethod
    def check_pointwise(
        self,
        data: np.ndarray[S, np.dtype[T]],
        prediction: np.ndarray[S, np.dtype[T]],
        *,
        late_bound: Bindings,
        where: Literal[True] | np.ndarray[S, np.dtype[np.bool]] = True,
    ) -> np.ndarray[S, np.dtype[np.bool]]:
        """
        Check which elements in the `prediction` array uphold the property
        enforced by this safeguard.

        Parameters
        ----------
        data : np.ndarray[S, np.dtype[T]]
            Original data array, relative to which the `prediction` is checked.
        prediction : np.ndarray[S, np.dtype[T]]
            Prediction for the `data` array.
        late_bound : Bindings
            Bindings for late-bound parameters, including for this safeguard.
        where : Literal[True] | np.ndarray[S, np.dtype[np.bool]]
            Only check at data points where the condition is [`True`][True].

        Returns
        -------
        ok : np.ndarray[S, np.dtype[np.bool]]
            Pointwise, `True` if the check succeeded for this element.
        """

        pass

    @abstractmethod
    def compute_safe_intervals(
        self,
        data: np.ndarray[S, np.dtype[T]],
        *,
        late_bound: Bindings,
        where: Literal[True] | np.ndarray[S, np.dtype[np.bool]] = True,
    ) -> IntervalUnion[T, int, int]:
        """
        Compute the intervals in which the safeguard's guarantees with respect
        to the `data` are upheld.

        The returned union of intervals must not have any overlap between the
        intervals inside the union. The `data` must be contained in the union.

        Parameters
        ----------
        data : np.ndarray[S, np.dtype[T]]
            Data for which the safe intervals should be computed.
        late_bound : Bindings
            Bindings for late-bound parameters, including for this safeguard.
        where : Literal[True] | np.ndarray[S, np.dtype[np.bool]]
            Only compute the safe intervals at data points where the condition
            is [`True`][True].

        Returns
        -------
        intervals : IntervalUnion[T, int, int]
            Union of intervals in which the safeguard's guarantees are upheld.
        """

        pass

    @abstractmethod
    def compute_footprint(
        self,
        foot: np.ndarray[S, np.dtype[np.bool]],
        *,
        late_bound: Bindings,
        where: Literal[True] | np.ndarray[S, np.dtype[np.bool]] = True,
    ) -> np.ndarray[S, np.dtype[np.bool]]:
        """
        Compute the footprint of the `foot` array, e.g. for expanding data
        points into the pointwise checks that they contribute to.

        For pointwise safeguards, the footprint should be equivalent to
        `foot & where`.

        Parameters
        ----------
        foot : np.ndarray[S, np.dtype[np.bool]]
            Array for which the footprint is computed.
        late_bound : Bindings
            Bindings for late-bound parameters, including for this safeguard.
        where : Literal[True] | np.ndarray[S, np.dtype[np.bool]]
            Only compute the footprint at pointwise checks where the condition
            is [`True`][True].

        Returns
        -------
        print : np.ndarray[S, np.dtype[np.bool]]
            The footprint of the `foot` array.
        """

        pass

    @abstractmethod
    def compute_inverse_footprint(
        self,
        foot: np.ndarray[S, np.dtype[np.bool]],
        *,
        late_bound: Bindings,
        where: Literal[True] | np.ndarray[S, np.dtype[np.bool]] = True,
    ) -> np.ndarray[S, np.dtype[np.bool]]:
        """
        Compute the inverse footprint of the `foot` array, e.g. for expanding
        pointwise check fails into the points that could have contributed to
        the failures.

        For pointwise safeguards, the inverse footprint should be equivalent to
        `foot & where`.

        Parameters
        ----------
        foot : np.ndarray[S, np.dtype[np.bool]]
            Array for which the inverse footprint is computed.
        late_bound : Bindings
            Bindings for late-bound parameters, including for this safeguard.
        where : Literal[True] | np.ndarray[S, np.dtype[np.bool]]
            Only compute the inverse footprint at pointwise checks where the
            condition is [`True`][True].

        Returns
        -------
        print : np.ndarray[S, np.dtype[np.bool]]
            The inverse footprint of the `foot` array.
        """

        pass



"""
Sign-preserving safeguard.
"""

__all__ = ["SignPreservingSafeguard"]

from collections.abc import Set
from typing import ClassVar, Literal

import numpy as np
from typing_extensions import override  # MSPV 3.12

from ...utils._compat import _ensure_array, _logical_and
from ...utils.bindings import Bindings, Parameter
from ...utils.cast import from_total_order, lossless_cast, to_total_order
from ...utils.error import TypeCheckError, ctx
from ...utils.intervals import Interval, IntervalUnion, Lower, Upper
from ...utils.typing import JSON, S, T
from .abc import PointwiseSafeguard


class SignPreservingSafeguard(PointwiseSafeguard):
    r"""
    The `SignPreservingSafeguard` guarantees that values have the same sign
    (-1, 0, +1) in the decompressed output as they have in the input data.

    NaN values are preserved as NaN values with the same sign bit.

    This safeguard can be configured to preserve the sign relative to a custom
    `offset`, e.g. to preserve global minima and maxima. The sign is then
    defined based on arithmetic comparison, i.e.

    \[
        \text{sign}(x, offset) = \begin{cases}
            -\text{NaN} \quad &\text{if } x \text{ is} -\text{NaN} \\
            -1 \quad &\text{if } x < offset \\
            0 \quad &\text{if } x = offset \\
            +1 \quad &\text{if } x > offset \\
            +\text{NaN} \quad &\text{if } x \text{ is} +\text{NaN}
        \end{cases}
    \]

    This safeguard should be combined with e.g. an error bound, as it by itself
    accepts *any* value with the same sign.

    Parameters
    ----------
    offset : int | float | str | Parameter, optional
        The non-NaN value of or the late-bound parameter name for the offset
        compared to which the sign is computed. By default, the offset is
        zero. Values that are above / below / equal to the offset are
        guaranteed to stay above / below / equal to the offset, respectively.
        Literal values are (unsafely) cast to the data dtype before comparison.

    Raises
    ------
    TypeCheckError
        if any parameter has the wrong type.
    ValueError
        if `offset` is NaN.
    """

    __slots__: tuple[str, ...] = ("_offset",)
    _offset: int | float | Parameter

    kind: ClassVar[str] = "sign"

    def __init__(self, *, offset: int | float | str | Parameter = 0) -> None:
        with ctx.safeguard(self):
            with ctx.parameter("offset"):
                TypeCheckError.check_instance_or_raise(
                    offset, int | float | str | Parameter
                )
                if isinstance(offset, Parameter):
                    self._offset = offset
                elif isinstance(offset, str):
                    self._offset = Parameter(offset)
                elif isinstance(offset, int) or (not np.isnan(offset)):
                    self._offset = offset
                else:
                    raise ValueError("must not be NaN") | ctx

    @property
    @override
    def late_bound(self) -> Set[Parameter]:
        """
        The set of late-bound parameters that this safeguard has.

        Late-bound parameters are only bound when checking and applying the
        safeguard, in contrast to the normal early-bound parameters that are
        configured during safeguard initialisation.

        Late-bound parameters can be used for parameters that depend on the
        specific data that is to be safeguarded.
        """

        return (
            frozenset([self._offset])
            if isinstance(self._offset, Parameter)
            else frozenset()
        )

    @override
    def check_pointwise(
        self,
        data: np.ndarray[S, np.dtype[T]],
        prediction: np.ndarray[S, np.dtype[T]],
        *,
        late_bound: Bindings,
        where: Literal[True] | np.ndarray[S, np.dtype[np.bool]] = True,
    ) -> np.ndarray[S, np.dtype[np.bool]]:
        """
        Check for which elements in the `prediction` array the signs match the
        signs of the `data` array elements'.

        Parameters
        ----------
        data : np.ndarray[S, np.dtype[T]]
            Original data array, relative to which the `prediction` is checked.
        prediction : np.ndarray[S, np.dtype[T]]
            Prediction for the `data` array.
        late_bound : Bindings
            Bindings for late-bound parameters, including for this safeguard.
        where : Literal[True] | np.ndarray[S, np.dtype[np.bool]]
            Only check at data points where the condition is [`True`][True].

        Returns
        -------
        ok : np.ndarray[S, np.dtype[np.bool]]
            Pointwise, `True` if the check succeeded for this element.

        Raises
        ------
        LateBoundParameterResolutionError
            if the `offset` is late-bound but its late-bound parameter is not in
            `late_bound`.
        ValueError
            if the late-bound `offset` could not be broadcast to the `data`'s
            shape.
        ValueError
            if the late-bound `offset` contains NaN values.
        TypeError
            if the `offset` is floating-point but the `data` is integer.
        ValueError
            if not all `offset`s could not be losslessly converted to the
            `data`'s type.
        """

        with ctx.safeguard(self), ctx.parameter("offset"):
            offset: np.ndarray[tuple[()] | S, np.dtype[T]] = (
                late_bound.resolve_ndarray_with_lossless_cast(
                    self._offset,
                    data.shape,
                    data.dtype,
                )
                if isinstance(self._offset, Parameter)
                else lossless_cast(self._offset, data.dtype)
            )
            if isinstance(self._offset, Parameter):
                if np.any(np.isnan(offset)):
                    with ctx.late_bound_parameter(self._offset):
                        raise ValueError("must not contain any NaN values") | ctx

        # values equal to the offset (sign=0) stay equal
        # values below (sign=-1) stay below,
        # values above (sign=+1) stay above
        # NaN values keep their sign bit
        ok: np.ndarray[S, np.dtype[np.bool]] = _ensure_array(prediction == offset)
        np.less(prediction, offset, out=ok, where=np.less(data, offset))
        np.greater(prediction, offset, out=ok, where=np.greater(data, offset))
        np.copyto(
            ok,
            np.isnan(prediction) & (np.signbit(data) == np.signbit(prediction)),
            where=np.isnan(data),
            casting="no",
        )
        if where is not True:
            ok[~where] = True

        return ok

    @override
    def compute_safe_intervals(
        self,
        data: np.ndarray[S, np.dtype[T]],
        *,
        late_bound: Bindings,
        where: Literal[True] | np.ndarray[S, np.dtype[np.bool]] = True,
    ) -> IntervalUnion[T, int, int]:
        """
        Compute the intervals in which the `data`'s sign is preserved.

        Parameters
        ----------
        data : np.ndarray[S, np.dtype[T]]
            Data for which the safe intervals should be computed.
        late_bound : Bindings
            Bindings for late-bound parameters, including for this safeguard.
        where : Literal[True] | np.ndarray[S, np.dtype[np.bool]]
            Only compute the safe intervals at pointwise checks where the
            condition is [`True`][True].

        Returns
        -------
        intervals : IntervalUnion[T, int, int]
            Union of intervals in which the `data`'s sign is preserved.

        Raises
        ------
        LateBoundParameterResolutionError
            if the `offset` is late-bound but its late-bound parameter is not
            in `late_bound`.
        ValueError
            if the late-bound `offset` could not be broadcast to the `data`'s
            shape.
        ValueError
            if the late-bound `offset` contains NaN values.
        TypeError
            if the `offset` is floating-point but the `data` is integer.
        ValueError
            if not all `offset`s could not be losslessly converted to the
            `data`'s type.
        """

        with ctx.safeguard(self), ctx.parameter("offset"):
            offsetf: np.ndarray[tuple[()] | tuple[int], np.dtype[T]] = (
                late_bound.resolve_ndarray_with_lossless_cast(
                    self._offset,
                    data.shape,
                    data.dtype,
                ).flatten()
                if isinstance(self._offset, Parameter)
                else lossless_cast(self._offset, data.dtype)
            )
            if isinstance(self._offset, Parameter):
                if np.any(np.isnan(offsetf)):
                    with ctx.late_bound_parameter(self._offset):
                        raise ValueError("must not contain any NaN values") | ctx
        offsetf_total: np.ndarray[
            tuple[()] | tuple[int], np.dtype[np.unsignedinteger]
        ] = to_total_order(offsetf)

        with np.errstate(over="ignore", under="ignore"):
            if np.issubdtype(data.dtype, np.floating):
                # special case for floating-point -0.0 / +0.0 which both have
                #  zero sign and thus have weird below / above intervals
                smallest_subnormal = np.finfo(data.dtype).smallest_subnormal  # type: ignore
                below_upper = _ensure_array(
                    from_total_order(offsetf_total - 1, data.dtype)
                )
                below_upper[offsetf == 0] = -smallest_subnormal
                above_lower = _ensure_array(
                    from_total_order(offsetf_total + 1, data.dtype)
                )
                above_lower[offsetf == 0] = smallest_subnormal
            else:
                below_upper = _ensure_array(
                    from_total_order(offsetf_total - 1, data.dtype)
                )
                above_lower = _ensure_array(
                    from_total_order(offsetf_total + 1, data.dtype)
                )

        dataf = data.flatten()
        valid = (
            Interval.empty_like(dataf)
            .preserve_signed_nan(dataf, equal_nan=True)
            .preserve_non_nan(dataf)
        )

        # preserve zero-sign values exactly
        Lower(dataf) <= valid[dataf == offsetf] <= Upper(dataf)
        valid[dataf < offsetf] <= Upper(below_upper)
        Lower(above_lower) <= valid[dataf > offsetf]

        wheref: Literal[True] | np.ndarray[tuple[int], np.dtype[np.bool]] = (
            True if where is True else where.flatten()
        )

        return valid.preserve_only_where(wheref).into_union()

    @override
    def compute_footprint(
        self,
        foot: np.ndarray[S, np.dtype[np.bool]],
        *,
        late_bound: Bindings,
        where: Literal[True] | np.ndarray[S, np.dtype[np.bool]] = True,
    ) -> np.ndarray[S, np.dtype[np.bool]]:
        """
        Compute the footprint of the `foot` array, e.g. for expanding data
        points into the pointwise checks that they contribute to.

        The footprint is equivalent to `foot & where`.

        Parameters
        ----------
        foot : np.ndarray[S, np.dtype[np.bool]]
            Array for which the footprint is computed.
        late_bound : Bindings
            Bindings for late-bound parameters, including for this safeguard.
        where : Literal[True] | np.ndarray[S, np.dtype[np.bool]]
            Only compute the footprint at pointwise checks where the condition
            is [`True`][True].

        Returns
        -------
        print : np.ndarray[S, np.dtype[np.bool]]
            The footprint of the `foot` array.
        """

        return _logical_and(foot, where)

    @override
    def compute_inverse_footprint(
        self,
        foot: np.ndarray[S, np.dtype[np.bool]],
        *,
        late_bound: Bindings,
        where: Literal[True] | np.ndarray[S, np.dtype[np.bool]] = True,
    ) -> np.ndarray[S, np.dtype[np.bool]]:
        """
        Compute the inverse footprint of the `foot` array, e.g. for expanding
        pointwise check fails into the points that could have contributed to
        the failures.

        The inverse footprint is equivalent to `foot & where`.

        Parameters
        ----------
        foot : np.ndarray[S, np.dtype[np.bool]]
            Array for which the inverse footprint is computed.
        late_bound : Bindings
            Bindings for late-bound parameters, including for this safeguard.
        where : Literal[True] | np.ndarray[S, np.dtype[np.bool]]
            Only compute the inverse footprint at pointwise checks where the
            condition is [`True`][True].

        Returns
        -------
        print : np.ndarray[S, np.dtype[np.bool]]
            The inverse footprint of the `foot` array.
        """

        return _logical_and(foot, where)

    @override
    def get_config(self) -> dict[str, JSON]:
        """
        Returns the configuration of the safeguard.

        Returns
        -------
        config : dict[str, JSON]
            Configuration of the safeguard.
        """

        return dict(kind=type(self).kind, offset=self._offset)



"""
Types and helpers for late-bound safeguard parameters.
"""

__all__ = ["Parameter", "Value", "Bindings"]

from base64 import standard_b64decode, standard_b64encode
from collections.abc import Set
from io import BytesIO
from types import MappingProxyType
from typing import Self, TypeAlias

import numpy as np
from typing_extensions import override  # MSPV 3.12

from ._compat import _broadcast_to
from .cast import lossless_cast, saturating_finite_float_cast
from .error import LateBoundParameterResolutionError, TypeCheckError, ctx
from .typing import JSON, F, Si, T


class Parameter(str):
    """
    Parameter name / identifier type.

    Parameters
    ----------
    param : str
        Name of the parameter, which must be a valid identifier.

    Raises
    ------
    ValueError
        if `param` is not a valid identifier.
    """

    __slots__: tuple[str, ...] = ()

    def __init__(self, param: str) -> None:
        pass

    def __new__(cls, param: str) -> "Parameter":
        if isinstance(param, Parameter):
            return param
        if not param.removeprefix("$").isidentifier():
            raise ValueError(f"parameter `{param}` is not a valid identifier") | ctx
        return super().__new__(cls, param)

    @property
    def is_builtin(self) -> bool:
        """
        Is the parameter a built-in parameter (starts with an `$`)?
        """

        return self.startswith("$")

    @override
    def __repr__(self) -> str:
        return f"{type(self).__name__}({self})"


Value: TypeAlias = (
    int | float | np.number | np.ndarray[tuple[int, ...], np.dtype[np.number]]
)
"""
Parameter value type that includes scalar numbers and arrays thereof.
"""


class Bindings:
    """
    Bindings from parameter names to values.

    Parameters
    ----------
    **kwargs : Value
        Mapping from parameters to values as keyword arguments.
    """

    __slots__: tuple[str, ...] = ("_bindings",)
    _bindings: MappingProxyType[Parameter, Value]

    EMPTY: "Bindings"
    """
    Empty bindings.
    """

    def __init__(self, **kwargs: Value) -> None:
        self._bindings = MappingProxyType(
            {Parameter(name): value for name, value in kwargs.items()}
        )

    def update(self, **kwargs: Value) -> Self:
        """
        Create new bindings that contain the old and the new parameters, where
        new parameters may override old ones.

        Parameters
        ----------
        **kwargs : Value
            Mapping from new parameters to values as keyword arguments.

        Returns
        -------
        bindings : Bindings
            The updated bindings.
        """

        return type(self)(**{str(p): v for p, v in self._bindings.items()}, **kwargs)

    def __contains__(self, param: Parameter) -> bool:
        """
        Checks if the `param`eter is contained in the bindings.

        Parameters
        ----------
        param : Parameter
            The parameter for which to check.

        Returns
        -------
        found : bool
            [`True`][True] if the bindings contain the `param`eter,
            [`False`][False] otherwise.
        """
        return param in self._bindings

    def parameters(self) -> Set[Parameter]:
        """
        Access the set of parameter names for which bindings exist.

        Returns
        -------
        params : Set[Parameter]
            The set of parameters in these bindings.
        """

        return self._bindings.keys()

    def resolve_ndarray_with_lossless_cast(
        self, param: Parameter, shape: Si, dtype: np.dtype[T]
    ) -> np.ndarray[Si, np.dtype[T]]:
        """
        Resolve the `param`eter to a numpy array with the given `shape` and
        `dtype`.

        The `param`eter must be contained in the bindings and refer to a value
        that can be broadcast to the `shape` and losslessly converted to the
        `dtype`.

        Parameters
        ----------
        param : Parameter
            The parameter that will be resolved.
        shape : Si
            The shape of the array to resolve to.
        dtype : np.dtype[T]
            The dtype of the array to resolve to.

        Returns
        -------
        array : np.ndarray[Si, np.dtype[T]]
            A read-only view to the resolved array of the given `shape` and
            `dtype`.

        Raises
        ------
        LateBoundParameterResolutionError
            if the `param`eter is not contained in the bindings.
        ValueError
            if the `param`eter could not be broadcast to the `shape`.
        TypeError
            if floating-point values are converted to an integer `dtype`.
        ValueError
            if not all values could be losslessly converted to `dtype`.
        """

        if param not in self._bindings:
            raise (
                LateBoundParameterResolutionError(frozenset([param]), frozenset([]))
                | ctx
            )

        # cast first then broadcast to allow zero-copy broadcasts of scalars
        #  to arrays of any shape
        with ctx.late_bound_parameter(param):
            value: np.ndarray[tuple[int, ...], np.dtype[T]] = lossless_cast(
                self._bindings[param], dtype
            )

            try:
                value_view: np.ndarray[Si, np.dtype[T]] = _broadcast_to(
                    value, shape
                ).view()
            except ValueError:
                raise (
                    ValueError(
                        f"cannot broadcast from shape {value.shape} to shape "
                        + f"{shape}"
                    )
                    | ctx
                ) from None

        value_view.flags.writeable = False

        return value_view

    def resolve_ndarray_with_saturating_finite_float_cast(
        self, param: Parameter, shape: Si, dtype: np.dtype[F]
    ) -> np.ndarray[Si, np.dtype[F]]:
        """
        Resolve the `param`eter to a numpy array with the given `shape` and
        floating-point `dtype`.

        The `param`eter must be contained in the bindings and refer to a finite
        value that can be broadcast to the `shape`. It will be converted to the
        floating-point `dtype`, with under- and overflows being clamped to
        finite values.

        Parameters
        ----------
        param : Parameter
            The parameter that will be resolved.
        shape : Si
            The shape of the array to resolve to.
        dtype : np.dtype[F]
            The floating-point dtype of the array to resolve to.

        Returns
        -------
        array : np.ndarray[Si, np.dtype[F]]
            A read-only view to the resolved array of the given `shape` and
            `dtype`.

        Raises
        ------
        LateBoundParameterResolutionError
            if the `param`eter is not contained in the bindings.
        ValueError
            if the `param`eter could not be broadcast to the `shape`.
        ValueError
            if any values are non-finite, i.e. infinite or NaN.
        """

        if param not in self._bindings:
            raise (
                LateBoundParameterResolutionError(frozenset([param]), frozenset([]))
                | ctx
            )

        # cast first then broadcast to allow zero-copy broadcasts of scalars
        #  to arrays of any shape
        with ctx.late_bound_parameter(param):
            value: np.ndarray[tuple[int, ...], np.dtype[F]] = (
                saturating_finite_float_cast(self._bindings[param], dtype)
            )

            try:
                value_view: np.ndarray[Si, np.dtype[F]] = _broadcast_to(
                    value, shape
                ).view()
            except ValueError:
                raise (
                    ValueError(
                        f"cannot broadcast from shape {value.shape} to shape "
                        + f"{shape}"
                    )
                    | ctx
                ) from None

        value_view.flags.writeable = False

        return value_view

    def expect_broadcastable_to(self, shape: tuple[int, ...]):
        """
        Check that all late-bound array parameters can be broadcast to the
        given `shape`.

        Parameters
        ----------
        shape : tuple[int, ...]
            The shape that all array parameters should be broadcastable to.

        Raises
        ------
        ValueError
            if an array parameter could not be broadcast to the `shape`.
        """

        for param, value in self._bindings.items():
            if isinstance(value, int | float | np.number):
                # scalar
                continue

            if not isinstance(value, np.ndarray):
                # bail out
                continue

            if value.ndim == 0:
                # scalar array
                continue

            with ctx.late_bound_parameter(param):
                if value.ndim != len(shape):
                    raise (
                        ValueError(
                            f"incompatible dimension {value.ndim}, expected "
                            + f"{len(shape)}"
                        )
                        | ctx
                    )

                if not all((v == 1) or (v == s) for v, s in zip(value.shape, shape)):
                    raise (
                        ValueError(
                            f"incompatible shape {value.shape}, expected {shape}"
                        )
                        | ctx
                    )

    def apply_slice_index(self, index: tuple[slice, ...]) -> Self:
        """
        Apply the slice `index` to the late-bound array values and return the sliced bindings.

        The `index` is only applied to an array value if
        - the value is not scalar (no effect)
        - the value has the same number of dimensions as the `index` (bail out)

        Furthermore, the index is only applied along dimensions with a size
        greater than 1, since smaller dimensions can be broadcast.

        Parameters
        ----------
        index : tuple[slice, ...]
            The slice index that is applied to the late-bound array values.

        Returns
        -------
        bindings : Bindings
            The sliced bindings.

        Raises
        ------
        IndexError
            if an array value has the right number of dimensions but indexing
            with `index` fails
        """

        def apply_index_to_value(value: Value) -> Value:
            if isinstance(value, int | float | np.number):
                # scalar
                return value

            if not isinstance(value, np.ndarray):
                # bail out
                return value

            if value.ndim == 0:
                # scalar array
                return value

            if value.ndim != len(index):
                # bail out
                return value

            # only apply the index along axes that cannot be broadcast
            value_index = tuple(
                i if a > 1 else slice(None, None) for i, a in zip(index, value.shape)
            )
            return value[value_index]

        return type(self)(
            **{
                str(param): apply_index_to_value(value)
                for param, value in self._bindings.items()
            }
        )

    def apply_roll(self, shift: tuple[int, ...]) -> Self:
        """
        Apply the roll `shift` to the late-bound array values and return the rolled bindings.

        The `shift` is only applied to an array value if
        - the value is not scalar (no effect)
        - the value has the same number of dimensions as the `shift` (bail out)

        Furthermore, the shift is only applied along dimensions with a size
        greater than 1, since smaller dimensions can be broadcast.

        Parameters
        ----------
        shift : tuple[int, ...]
            The roll shift that is applied to the late-bound array values.

        Returns
        -------
        bindings : Bindings
            The rolled bindings.
        """

        def apply_roll_to_value(value: Value) -> Value:
            if isinstance(value, int | float | np.number):
                # scalar
                return value

            if not isinstance(value, np.ndarray):
                # bail out
                return value

            if value.ndim == 0:
                # scalar array
                return value

            if value.ndim != len(shift):
                # bail out
                return value

            # only apply the roll along axes that cannot be broadcast
            value_shift = tuple(s if a > 1 else 0 for s, a in zip(shift, value.shape))
            return np.roll(value, value_shift, axis=tuple(range(value.ndim)))

        return type(self)(
            **{
                str(param): apply_roll_to_value(value)
                for param, value in self._bindings.items()
            }
        )

    def get_config(self) -> dict[str, JSON]:
        """
        Returns the configuration of the bindings in a JSON compatible format.

        Returns
        -------
        config : dict[str, JSON]
            Configuration of the bindings.
        """

        return {str(p): _encode_value(p, v) for p, v in self._bindings.items()}

    @classmethod
    def from_config(cls, config: dict[str, JSON]) -> Self:
        """
        Instantiate the bindings from a configuration [`dict`][dict].

        Parameters
        ----------
        config : dict[str, JSON]
            Configuration of the bindings.

        Returns
        -------
        bindings : Self
            Instantiated bindings.
        """

        return cls(**{p: _decode_value(Parameter(p), v) for p, v in config.items()})  # type: ignore


Bindings.EMPTY = Bindings()


_NPZ_DATA_URI_BASE64: str = "data:application/x-npz;base64,"


def _encode_value(p: Parameter, v: Value) -> int | float | str:
    # we cannot use isinstance here since np.float64 is a subclass of float
    if type(v) is int:
        return v
    if type(v) is float:
        return v

    io = BytesIO()
    np.savez_compressed(io, allow_pickle=False, **{str(p): v})

    return f"{_NPZ_DATA_URI_BASE64}{standard_b64encode(io.getvalue()).decode(encoding='ascii')}"


def _decode_value(p: Parameter, v: int | float | str) -> Value:
    TypeCheckError.check_instance_or_raise(v, int | float | str)

    if isinstance(v, int | float):
        return v

    if not v.startswith(_NPZ_DATA_URI_BASE64):
        with ctx.parameter(p):
            raise (
                ValueError("must be encoded as a .npz data URI in base64 format") | ctx
            )

    io = BytesIO(
        standard_b64decode(v[len(_NPZ_DATA_URI_BASE64) :].encode(encoding="ascii"))
    )

    with np.load(io, allow_pickle=False) as f:
        return f[str(p)]



"""
Types and helpers to construct safe intervals for the safeguards.
"""

__all__ = [
    "Interval",
    "IntervalUnion",
    "Minimum",
    "Maximum",
    "Lower",
    "Upper",
    "N",
    "Ni",
    "U",
    "Ui",
]

from typing import Any, Generic, Literal, Self, TypeVar

import numpy as np
from typing_extensions import override  # MSPV 3.12

from ._compat import _ensure_array, _reshape, _where, _zeros
from ._compat import _maximum_zero_sign_sensitive as _np_maximum
from ._compat import _minimum_zero_sign_sensitive as _np_minimum
from .cast import as_bits, from_total_order, to_total_order
from .error import TypeSetError
from .typing import S, T

N = TypeVar("N", bound=int, covariant=True)
""" The number of elements in the interval (covariant). """

Ni = TypeVar("Ni", bound=int)
""" The number of elements in the interval (invariant). """

U = TypeVar("U", bound=int, covariant=True)
""" The maximum number of intervals in an interval union (covariant). """

Ui = TypeVar("Ui", bound=int)
""" The maximum number of intervals in an interval union (invariant). """


class Interval(Generic[T, N]):
    """
    Single interval over a `N`-sized [`ndarray`][numpy.ndarray] of data type `T`.

    ## Construction

    An interval should only be constructed using

    - [`Interval.empty`][.empty] or [`Interval.empty_like`][.empty_like] for an
      empty interval that contains no values
    - [`Interval.full`][.full] or [`Interval.full_like`][.full_like] for a
      full interval that contains all possible values

    ## Overriding lower and upper bounds

    The lower and upper bounds of the interval can be set as follows:

    ```python
    Lower(lower_data) <= interval <= Upper(upper_data)
    ```

    They can also be set to the smallest and largest possible values using:

    ```python
    Minimum <= interval <= Maximum
    ```

    If only the interval bounds for some array members should be updated, only
    the interval itself needs to be indexed:

    ```python
    # bound infinite values to be exactly themselves
    Lower(data) <= interval[np.isinf(data)] <= Upper(data)
    ```

    ## Common lower and upper bounds

    The following common lower and upper bounds are provided for ease of use:

    - [`Interval.preserve_inf`][.preserve_inf]
    - [`Interval.preserve_signed_nan`][.preserve_signed_nan]
    - [`Interval.preserve_any_nan`][.preserve_any_nan]
    - [`Interval.preserve_finite`][.preserve_finite]
    - [`Interval.preserve_non_nan`][.preserve_non_nan]

    ## Interval operations

    Two intervals can be

    - intersected using [`Interval.intersect`][.intersect]
    - unioned using [`Interval.union`][.union]

    or converted into a single-member union of intervals using
    [`Interval.into_union`][.into_union].
    """

    __slots__: tuple[str, ...] = ("_lower", "_upper")
    _lower: np.ndarray[tuple[N], np.dtype[T]]
    _upper: np.ndarray[tuple[N], np.dtype[T]]

    def __init__(
        self,
        *,
        _lower: np.ndarray[tuple[N], np.dtype[T]],
        _upper: np.ndarray[tuple[N], np.dtype[T]],
    ) -> None:
        self._lower = _lower
        self._upper = _upper

    @property
    def dtype(self) -> np.dtype[T]:
        """
        Dtype `T` of the interval.
        """
        return self._lower.dtype

    @property
    def n(self) -> N:
        """
        Size `N` of the interval.
        """
        (n,) = self._lower.shape
        return n

    def non_empty_width(self) -> np.ndarray[tuple[N], np.dtype[np.unsignedinteger]]:
        """
        Binary width of the interval.

        The interval must not have any empty entries.

        Single-element intervals have width zero, full intervals have maximum
        width.

        Returns
        -------
        widths : np.ndarray[tuple[N], np.dtype[np.unsignedinteger]]
            The binary widths of the intervals.
        """

        lt: np.ndarray[tuple[N], np.dtype[np.unsignedinteger]] = to_total_order(
            self._lower
        )
        ut: np.ndarray[tuple[N], np.dtype[np.unsignedinteger]] = to_total_order(
            self._upper
        )

        assert np.all(lt <= ut)

        return ut - lt

    @staticmethod
    def empty(dtype: np.dtype[T], n: Ni) -> "Interval[T, Ni]":
        """
        Create an empty interval that contains no values.

        Parameters
        ----------
        dtype : np.dtype[T]
            The dtype of the interval
        n : Ni
            The size of the interval

        Returns
        -------
        empty : Interval[T, Ni]
            The empty interval
        """

        single: IntervalUnion[T, Ni, Literal[1]] = IntervalUnion.empty(dtype, n, 1)
        return Interval(
            _lower=single._lower[0],
            _upper=single._upper[0],
        )

    @staticmethod
    def empty_like(a: np.ndarray[tuple[int, ...], np.dtype[T]]) -> "Interval[T, int]":
        """
        Create an empty interval that contains no values and has the same dtype and size as `a`.

        Parameters
        ----------
        a : np.ndarray[tuple[int, ...], np.dtype[T]]
            An array whose dtype and size the interval gets.

        Returns
        -------
        empty : Interval[T, int]
            The empty interval
        """

        return Interval.empty(a.dtype, a.size)

    @staticmethod
    def full(dtype: np.dtype[T], n: Ni) -> "Interval[T, Ni]":
        """
        Create a full interval that contains all possible values of `dtype`.

        Parameters
        ----------
        dtype : np.dtype[T]
            The dtype of the interval
        n : Ni
            The size of the interval

        Returns
        -------
        full : Interval[T, Ni]
            The full interval
        """

        return Interval(
            _lower=np.full(n, _minimum(dtype), dtype=dtype),
            _upper=np.full(n, _maximum(dtype), dtype=dtype),
        )

    @staticmethod
    def full_like(a: np.ndarray[tuple[int, ...], np.dtype[T]]) -> "Interval[T, int]":
        """
        Create a full interval that contains all possible values and has the same dtype and size as `a`.

        Parameters
        ----------
        a : np.ndarray[tuple[int, ...], np.dtype[T]]
            An array whose dtype and size the interval gets

        Returns
        -------
        full : Interval[T, int]
            The full interval
        """

        return Interval.full(a.dtype, a.size)

    def __getitem__(self, key) -> "IndexedInterval[T, N]":
        return IndexedInterval(_lower=self._lower, _upper=self._upper, _index=key)

    def preserve_inf(self, a: np.ndarray[tuple[N], np.dtype[T]]) -> Self:
        """
        Preserve all infinite values in `a` exactly.

        Specifically, set their lower and upper bounds in this interval to
        their values.

        Equivalent to

        ```python
        Lower(a) <= self[np.isinf(a)] <= Upper(a)
        ```

        Parameters
        ----------
        a : np.ndarray[tuple[N], np.dtype[T]]
            The arrays whose infinite values this interval should preserve

        Returns
        -------
        self : Self
            The modified `self`.
        """

        if not np.issubdtype(a.dtype, np.floating):
            return self

        # bitwise preserve infinite values
        Lower(a) <= self[np.isinf(a)] <= Upper(a)

        return self

    def preserve_signed_nan(
        self, a: np.ndarray[tuple[N], np.dtype[T]], *, equal_nan: bool
    ) -> Self:
        """
        Preserve all NaN values in `a`, preserving their sign bit.

        - If `equal_nan` is [`True`][True], the intervals corresponding to
          the NaN values will include all possible NaN values with the same
          sign bit.
        - If `equal_nan` is [`False`][False], all NaN values are preserved
          exactly.

        Parameters
        ----------
        a : np.ndarray[tuple[N], np.dtype[T]]
            The arrays whose NaN values this interval should preserve
        equal_nan : bool
            Whether any NaN values matches another NaN value or if NaN values
            should be preserved exactly

        Returns
        -------
        self : Self
            The modified `self`.
        """

        if not np.issubdtype(a.dtype, np.floating):
            return self

        if not equal_nan:
            # bitwise preserve NaN values
            Lower(a) <= self[np.isnan(a)] <= Upper(a)
            return self

        # smallest (positive) NaN bit pattern: 0b s 1..1 0..0
        nan_min = np.array(as_bits(np.array(np.inf, dtype=a.dtype)) + 1).view(a.dtype)
        # largest (negative) NaN bit pattern: 0b s 1..1 1..1
        nan_max = np.array(-1, dtype=a.dtype.str.replace("f", "i")).view(a.dtype)

        # any NaN with the same sign is valid
        # this is slightly stricter than what equal_nan requires
        Lower(
            _where(
                # ensure the NaN has the correct sign
                np.signbit(a),
                np.copysign(nan_max, -1),
                np.copysign(nan_min, +1),
            )
        ) <= self[np.isnan(a)] <= Upper(
            _where(
                np.signbit(a),
                np.copysign(nan_min, -1),
                np.copysign(nan_max, +1),
            )
        )

        return self

    def preserve_any_nan(
        self, a: np.ndarray[tuple[N], np.dtype[T]], *, equal_nan: bool
    ) -> "IntervalUnion[T, N, int]":
        """
        Preserve all NaN values in `a`, ignoring their sign bit.

        - If `equal_nan` is [`True`][True], the intervals corresponding to
          the NaN values will include all possible NaN values, irrespective of
          their sign bit.
        - If `equal_nan` is [`False`][False], all NaN values are preserved
          exactly.

        Since there are two disjoint value regions of NaNs (positive and
        negative), this method returns a union of intervals.

        Parameters
        ----------
        a : np.ndarray[tuple[N], np.dtype[T]]
            The arrays whose NaN values this interval should preserve
        equal_nan : bool
            Whether any NaN values matches another NaN value or if NaN values
            should be preserved exactly

        Returns
        -------
        union : IntervalUnion[T, N, int]
            The union of the existing intervals for non-NaN values and the
            NaN-preserving intervals for NaN values.
        """

        if not np.issubdtype(a.dtype, np.floating):
            return self.into_union()

        (n,) = a.shape

        lower: Interval[T, N] = Interval.empty(a.dtype, n)
        # copy over the intervals for non-NaN elements
        Lower(self._lower) <= lower[~np.isnan(a)] <= Upper(self._upper)

        if (not equal_nan) or (not np.any(np.isnan(a))):
            # bitwise preserve NaN values
            Lower(a) <= lower[np.isnan(a)] <= Upper(a)
            return lower.into_union()

        # smallest (positive) NaN bit pattern: 0b s 1..1 0..0
        nan_min = np.array(as_bits(np.array(np.inf, dtype=a.dtype)) + 1).view(a.dtype)
        # largest (negative) NaN bit pattern: 0b s 1..1 1..1
        nan_max = np.array(-1, dtype=a.dtype.str.replace("f", "i")).view(a.dtype)

        upper: Interval[T, N] = Interval.empty(a.dtype, n)

        # create lower interval of all negative NaNs
        Lower(np.array(np.copysign(nan_max, -1))) <= lower[np.isnan(a)] <= Upper(
            np.array(np.copysign(nan_min, -1))
        )
        # create upper interval of all positive NaNs
        Lower(np.array(np.copysign(nan_min, +1))) <= upper[np.isnan(a)] <= Upper(
            np.array(np.copysign(nan_max, +1))
        )

        return lower.union(upper)

    def preserve_finite(self, a: np.ndarray[tuple[N], np.dtype[T]]) -> Self:
        """
        Preserve all finite values in `a` as finite values.

        Specifically, set their lower and upper bounds to exclude non-finite
        values.

        Parameters
        ----------
        a : np.ndarray[tuple[N], np.dtype[T]]
            The arrays whose non-finite values this interval should preserve as
            non-finite.

        Returns
        -------
        self : Self
            The modified `self`.
        """

        if not np.issubdtype(a.dtype, np.floating):
            Minimum <= self[:] <= Maximum
            return self

        # nextafter produces the largest and smallest finite floating-point
        #  values
        Lower(
            np.array(
                np.nextafter(
                    a.dtype.type(-np.inf),  # type: ignore
                    a.dtype.type(0),
                )
            )
        ) <= self[np.isfinite(a)] <= Upper(
            np.array(
                np.nextafter(
                    a.dtype.type(np.inf),  # type: ignore
                    a.dtype.type(0),
                )
            )
        )

        return self

    def preserve_non_nan(self, a: np.ndarray[tuple[N], np.dtype[T]]) -> Self:
        """
        Preserve all non-NaN values in `a` as non-NaN values.

        Specifically, set their lower and upper bounds to exclude NaN values.

        Parameters
        ----------
        a : np.ndarray[tuple[N], np.dtype[T]]
            The arrays whose non-NaN values this interval should preserve as
            non-NaN.

        Returns
        -------
        self : Self
            The modified `self`.
        """

        if not np.issubdtype(a.dtype, np.floating):
            Minimum <= self[:] <= Maximum
            return self

        Lower(np.array(-np.inf, dtype=a.dtype)) <= self[~np.isnan(a)] <= Upper(
            np.array(np.inf, dtype=a.dtype)
        )

        return self

    def intersect(self, other: "Interval[T, N]") -> "Interval[T, N]":
        """
        Computes the intersection with the `other` interval.

        Parameters
        ----------
        other : Interval[T, N]
            The other interval to intersect with

        Returns
        -------
        intersection : Interval[T, N]
            The intersection of `self` and `other`.
        """

        (n,) = self._lower.shape

        if n == 0:
            return Interval.empty(self._lower.dtype, n)

        out: Interval[T, N] = Interval.empty(
            self._lower.dtype,
            n,
        )

        intersection_lower: np.ndarray[tuple[N], np.dtype[np.unsignedinteger]] = (
            _np_maximum(to_total_order(self._lower), to_total_order(other._lower))
        )
        intersection_upper: np.ndarray[tuple[N], np.dtype[np.unsignedinteger]] = (
            _np_minimum(to_total_order(self._upper), to_total_order(other._upper))
        )

        out._lower[:] = from_total_order(intersection_lower, out._lower.dtype)
        out._upper[:] = from_total_order(intersection_upper, out._upper.dtype)

        return out

    def union(
        self, other: "Interval[T, N] | IntervalUnion[T, N, int]"
    ) -> "IntervalUnion[T, N, int]":
        """
        Computes the union with the `other` interval (union).

        Parameters
        ----------
        other : Interval[T, N] | IntervalUnion[T, N, int]
            The other interval (union) to union with

        Returns
        -------
        intersection : IntervalUnion[T, N, int]
            The union of `self` and `other`.
        """

        return self.into_union().union(other)

    def into_union(self) -> "IntervalUnion[T, N, Literal[1]]":
        """
        Convert this interval into a single-member union of exactly this one interval.

        Returns
        -------
        union : IntervalUnion[T, N, Literal[1]]
            The union of only this interval
        """

        return IntervalUnion(
            _lower=_reshape(self._lower, (1, self.n)),
            _upper=_reshape(self._upper, (1, self.n)),
        )

    def preserve_only_where(
        self, where: Literal[True] | np.ndarray[tuple[N], np.dtype[np.bool]]
    ) -> Self:
        """
        Preserve safe intervals only at the data points where the `where`
        condition is [`True`][True].

        This method must only be used for pointwise safe intervals.

        Parameters
        ----------
        where : Literal[True] | np.ndarray[tuple[N], np.dtype[np.bool]]
            Only preserve safe intervals at data points where the condition is
            [`True`][True].

        Returns
        -------
        self : Self
            The modified `self`.
        """

        if where is True:
            return self

        # reset all elements where where is False to a full interval
        Minimum <= self[~where] <= Maximum
        return self

    @override
    def __repr__(self) -> str:
        return f"Interval(lower={self._lower!r}, upper={self._upper!r})"


class IndexedInterval(Generic[T, N]):
    __slots__: tuple[str, ...] = ("_lower", "_upper", "_index")
    _lower: np.ndarray[tuple[N], np.dtype[T]]
    _upper: np.ndarray[tuple[N], np.dtype[T]]
    _index: Any

    def __init__(
        self,
        *,
        _lower: np.ndarray[tuple[N], np.dtype[T]],
        _upper: np.ndarray[tuple[N], np.dtype[T]],
        _index: Any,
    ) -> None:
        self._lower = _lower
        self._upper = _upper
        self._index = _index


def _minimum(dtype: np.dtype[T]) -> np.ndarray[tuple[()], np.dtype[T]]:
    TypeSetError.check_or_raise(dtype.type, np.integer | np.floating)

    if np.issubdtype(dtype, np.integer):
        return np.array(np.iinfo(dtype).min, dtype=dtype)  # type: ignore

    btype = dtype.str.replace("f", "u")
    bmin = np.iinfo(btype).max  # produces -NaN (0xffff...)
    return np.array(bmin, dtype=btype).view(dtype)


class _Minimum:
    __slots__: tuple[str, ...] = ()

    def __le__(self, interval: IndexedInterval[T, N]) -> IndexedInterval[T, N]:
        if not isinstance(interval, IndexedInterval):
            return NotImplemented

        interval._lower[interval._index] = _minimum(interval._lower.dtype)

        return interval


Minimum = _Minimum()
""" The smallest representable value """


def _maximum(dtype: np.dtype[T]) -> np.ndarray[tuple[()], np.dtype[T]]:
    TypeSetError.check_or_raise(dtype.type, np.integer | np.floating)

    if np.issubdtype(dtype, np.integer):
        return np.array(np.iinfo(dtype).max, dtype=dtype)  # type: ignore

    btype = dtype.str.replace("f", "u")
    bmin = np.iinfo(btype).max  # produces -NaN (0xffff...)
    return np.copysign(np.array(bmin, dtype=btype).view(dtype), +1)  # type: ignore


class _Maximum:
    __slots__: tuple[str, ...] = ()

    def __ge__(self, interval: IndexedInterval[T, N]) -> IndexedInterval[T, N]:
        if not isinstance(interval, IndexedInterval):
            return NotImplemented

        interval._upper[interval._index] = _maximum(interval._upper.dtype)

        return interval


Maximum = _Maximum()
""" The largest representable value """


class Lower:
    """
    Array wrapper to override an [`Interval`][compression_safeguards.utils.intervals.Interval]'s lower bound using comparison syntax.

    ```python
    Lower(lower_bound) <= interval
    ```

    Parameters
    ----------
    lower : np.ndarray[tuple[int, ...], np.dtype[np.number]]
        The lower bound array
    """

    __slots__: tuple[str, ...] = ("_lower",)
    _lower: np.ndarray[tuple[int, ...], np.dtype[np.number]]

    def __init__(self, lower: np.ndarray[tuple[int, ...], np.dtype[np.number]]) -> None:
        self._lower = lower

    def __le__(self, interval: IndexedInterval[T, N]) -> IndexedInterval[T, N]:
        if not isinstance(self._lower, np.ndarray):
            return NotImplemented

        if not isinstance(interval, IndexedInterval):
            return NotImplemented

        if self._lower.dtype != interval._lower.dtype:
            return NotImplemented

        if self._lower.shape == ():
            lower = self._lower
        elif self._lower.shape == interval._lower.shape:
            lower = self._lower[interval._index]
        elif self._lower.shape == interval._lower[interval._index].shape:
            lower = self._lower
        else:
            return NotImplemented

        interval._lower[interval._index] = lower

        return interval


class Upper:
    """
    Array wrapper to override an [`Interval`][compression_safeguards.utils.intervals.Interval]'s upper bound using comparison syntax.

    ```python
    interval <= Upper(upper_bound)
    ```

    Parameters
    ----------
    upper : np.ndarray[tuple[int, ...], np.dtype[np.number]]
        The upper bound array
    """

    __slots__: tuple[str, ...] = ("_upper",)
    _upper: np.ndarray[tuple[int, ...], np.dtype[np.number]]

    def __init__(self, upper: np.ndarray[tuple[int, ...], np.dtype[np.number]]) -> None:
        self._upper = upper

    def __ge__(self, interval: IndexedInterval[T, N]) -> IndexedInterval[T, N]:
        if not isinstance(self._upper, np.ndarray):
            return NotImplemented

        if not isinstance(interval, IndexedInterval):
            return NotImplemented

        if self._upper.dtype != interval._upper.dtype:
            return NotImplemented

        if self._upper.shape == ():
            upper = self._upper
        elif self._upper.shape == interval._upper.shape:
            upper = self._upper[interval._index]
        elif self._upper.shape == interval._upper[interval._index].shape:
            upper = self._upper
        else:
            return NotImplemented

        interval._upper[interval._index] = upper

        return interval


class IntervalUnion(Generic[T, N, U]):
    """
    Union of `U` intervals, each over a `N`-sized [`ndarray`][numpy.ndarray] of data type `T`.
    """

    __slots__: tuple[str, ...] = ("_lower", "_upper")
    # invariants:
    # - the lower/upper bounds are in sorted order
    # - no non-empty intervals come after empty intervals
    # - no intervals intersect within the union
    # - no intervals can be adjacent within the union, e.g. [1..3] and [4..5]
    _lower: np.ndarray[tuple[U, N], np.dtype[T]]
    _upper: np.ndarray[tuple[U, N], np.dtype[T]]

    def __init__(
        self,
        *,
        _lower: np.ndarray[tuple[U, N], np.dtype[T]],
        _upper: np.ndarray[tuple[U, N], np.dtype[T]],
    ) -> None:
        self._lower = _lower
        self._upper = _upper

    @property
    def dtype(self) -> np.dtype[T]:
        """
        Dtype `T` of the interval union.
        """
        return self._lower.dtype

    @property
    def n(self) -> N:
        """
        Size `N` of the interval union.
        """
        u, n = self._lower.shape
        return n

    @property
    def u(self) -> U:
        """
        Number of intervals `U` in the interval union.
        """
        u, n = self._lower.shape
        return u

    def non_empty_width(self) -> np.ndarray[tuple[N], np.dtype[np.unsignedinteger]]:
        """
        Binary width of the interval union.

        The interval union must not have any empty entries.

        Single-element intervals have width zero, full intervals have maximum
        width.

        Returns
        -------
        widths : np.ndarray[tuple[N], np.dtype[np.unsignedinteger]]
            The binary widths of the intervals.
        """

        (u, n) = self._lower.shape
        interval_width_acc: np.ndarray[tuple[N], np.dtype[np.unsignedinteger]] = _zeros(
            (n,), dtype=to_total_order(np.array((), dtype=self._lower.dtype)).dtype
        )

        for i in range(u):
            lt: np.ndarray[tuple[N], np.dtype[np.unsignedinteger]] = to_total_order(
                self._lower[i]
            )
            ut: np.ndarray[tuple[N], np.dtype[np.unsignedinteger]] = to_total_order(
                self._upper[i]
            )

            if i == 0:
                assert np.all(lt <= ut)

            np.add(
                interval_width_acc,
                ut - lt + 1,
                out=interval_width_acc,
                where=(ut >= lt),
            )

        interval_width_acc -= 1

        return interval_width_acc

    @staticmethod
    def empty(dtype: np.dtype[T], n: Ni, u: Ui) -> "IntervalUnion[T, Ni, Ui]":
        """
        Create an empty interval union that contains no values.

        Parameters
        ----------
        dtype : np.dtype[T]
            The dtype of the intervals
        n : Ni
            The size of the intervals
        u : Ui
            The number of intervals in the union

        Returns
        -------
        empty : IntervalUnion[T, Ni, Ui]
            The empty interval union
        """

        return IntervalUnion(
            _lower=np.full((u, n), _maximum(dtype), dtype=dtype),
            _upper=np.full((u, n), _minimum(dtype), dtype=dtype),
        )

    def intersect(
        self, other: "IntervalUnion[T, N, int]"
    ) -> "IntervalUnion[T, N, int]":
        """
        Computes the intersection with the `other` interval union.

        Parameters
        ----------
        other : IntervalUnion[T, N, int]
            The other interval union to intersect with

        Returns
        -------
        intersection : IntervalUnion[T, N, int]
            The intersection of `self` and `other`.
        """

        ((u, n), (v, _)) = self._lower.shape, other._lower.shape

        if n == 0:
            return IntervalUnion.empty(self._lower.dtype, n, 0)

        uv: int = u + v - 1
        out: IntervalUnion[T, N, int] = IntervalUnion.empty(
            self._lower.dtype,
            n,
            uv,
        )
        n_intervals = np.zeros(n, dtype=int)

        for i in range(u):
            for j in range(v):
                intersection_lower: np.ndarray[
                    tuple[U, N], np.dtype[np.unsignedinteger]
                ] = _np_maximum(
                    to_total_order(self._lower[i]), to_total_order(other._lower[j])
                )
                intersection_upper: np.ndarray[
                    tuple[U, N], np.dtype[np.unsignedinteger]
                ] = _np_minimum(
                    to_total_order(self._upper[i]), to_total_order(other._upper[j])
                )

                has_intersection = intersection_lower <= intersection_upper

                out._lower[n_intervals[has_intersection], has_intersection] = (
                    from_total_order(
                        intersection_lower[has_intersection], out._lower.dtype
                    )
                )
                out._upper[n_intervals[has_intersection], has_intersection] = (
                    from_total_order(
                        intersection_upper[has_intersection], out._upper.dtype
                    )
                )

                n_intervals += has_intersection

        uv = np.amax(n_intervals)

        return IntervalUnion(_lower=out._lower[:uv], _upper=out._upper[:uv])  # type: ignore

    def union(
        self, other: "Interval[T, N] | IntervalUnion[T, N, int]"
    ) -> "IntervalUnion[T, N, int]":
        """
        Computes the union with the `other` interval (union).

        Parameters
        ----------
        other : Interval[T, N] | IntervalUnion[T, N, int]
            The other interval (union) to union with

        Returns
        -------
        union : IntervalUnion[T, N, int]
            The union of `self` and `other`.
        """

        otheru = other if isinstance(other, IntervalUnion) else other.into_union()

        ((u, n), (v, _)) = self._lower.shape, otheru._lower.shape

        if n == 0:
            return IntervalUnion.empty(self._lower.dtype, n, 0)

        if u == 0:
            return otheru

        if v == 0:
            return self

        uv: int = u + v
        out: IntervalUnion[T, N, int] = IntervalUnion.empty(
            self._lower.dtype,
            n,
            uv,
        )
        n_intervals = np.zeros(n, dtype=int)

        i_s = np.zeros(n, dtype=int)
        j_s = np.zeros(n, dtype=int)

        while (np.amin(i_s) < u) or (np.amin(j_s) < v):
            lower_i: np.ndarray[tuple[int], np.dtype[np.unsignedinteger]] = (
                to_total_order(
                    np.take_along_axis(
                        self._lower, _np_minimum(i_s, u - 1).reshape(1, -1), axis=0
                    ).flatten()
                )
            )
            upper_i: np.ndarray[tuple[int], np.dtype[np.unsignedinteger]] = (
                to_total_order(
                    np.take_along_axis(
                        self._upper, _np_minimum(i_s, u - 1).reshape(1, -1), axis=0
                    ).flatten()
                )
            )

            lower_j: np.ndarray[tuple[int], np.dtype[np.unsignedinteger]] = (
                to_total_order(
                    np.take_along_axis(
                        otheru._lower, _np_minimum(j_s, v - 1).reshape(1, -1), axis=0
                    ).flatten()
                )
            )
            upper_j: np.ndarray[tuple[int], np.dtype[np.unsignedinteger]] = (
                to_total_order(
                    np.take_along_axis(
                        otheru._upper, _np_minimum(j_s, v - 1).reshape(1, -1), axis=0
                    ).flatten()
                )
            )

            lower_o: np.ndarray[tuple[int], np.dtype[np.unsignedinteger]] = (
                to_total_order(
                    np.take_along_axis(
                        out._lower,
                        _np_maximum(n_intervals - 1, 0).reshape(1, -1),
                        axis=0,
                    ).flatten()
                )
            )
            upper_o: np.ndarray[tuple[int], np.dtype[np.unsignedinteger]] = (
                to_total_order(
                    np.take_along_axis(
                        out._upper,
                        _np_maximum(n_intervals - 1, 0).reshape(1, -1),
                        axis=0,
                    ).flatten()
                )
            )

            # only valid if in-bounds and non-empty
            valid_i = (i_s < u) & (lower_i <= upper_i)
            valid_j = (j_s < v) & (lower_j <= upper_j)
            valid_o = (n_intervals > 0) & (lower_o <= upper_o)

            # choose the next valid interval with the lower lower bound
            choose_i = valid_i & ((lower_i < lower_j) | ~valid_j)
            choose_j = valid_j & ((lower_i >= lower_j) | ~valid_i)

            lower_ij = _ensure_array(lower_j)
            np.copyto(lower_ij, lower_i, where=choose_i, casting="no")
            upper_ij = _ensure_array(upper_j)
            np.copyto(upper_ij, upper_i, where=choose_i, casting="no")

            # check if the selected interval intersects with the previously
            #  output interval
            has_intersection_with_out = (
                (valid_i | valid_j)
                & valid_o
                & (
                    # check for normal intersection
                    (_np_maximum(lower_ij, lower_o) <= _np_minimum(upper_ij, upper_o))
                    |
                    # check for adjacent intervals, e.g [1..3] | [4..5] -> [1..5]
                    ((lower_ij > upper_o) & (lower_ij == (upper_o + 1)))
                )
            )

            # - intersection -> next is intersection
            # - no intersection -> next is next interval
            next_lower = _ensure_array(lower_ij, copy=True)
            np.copyto(
                next_lower,
                _np_minimum(lower_o, lower_ij),
                where=has_intersection_with_out,
                casting="no",
            )

            next_upper = _ensure_array(upper_ij, copy=True)
            np.copyto(
                next_upper,
                _np_maximum(upper_o, upper_ij),
                where=has_intersection_with_out,
                casting="no",
            )

            # update either the previous or the next output interval
            has_next = has_intersection_with_out | choose_i | choose_j
            out._lower[
                n_intervals[has_next] - has_intersection_with_out[has_next],
                has_next,
            ] = from_total_order(next_lower[has_next], out._lower.dtype)
            out._upper[
                n_intervals[has_next] - has_intersection_with_out[has_next],
                has_next,
            ] = from_total_order(next_upper[has_next], out._upper.dtype)

            # advance to the next interval if we wrote out a new one
            n_intervals += has_next & (~has_intersection_with_out)

            # advance the interval that was chosen earlier
            i_s += choose_i | (~valid_j)
            j_s += choose_j | (~valid_i)

        uv = np.amax(n_intervals)

        return IntervalUnion(_lower=out._lower[:uv], _upper=out._upper[:uv])  # type: ignore

    def preserve_only_where(
        self, where: Literal[True] | np.ndarray[tuple[N], np.dtype[np.bool]]
    ) -> "IntervalUnion[T, N, int]":
        """
        Preserve safe intervals only at the data points where the `where`
        condition is [`True`][True].

        This method must only be used for pointwise safe intervals.

        Parameters
        ----------
        where : Literal[True] | np.ndarray[tuple[N], np.dtype[np.bool]]
            Only preserve safe intervals at data points where the condition is
            [`True`][True].

        Returns
        -------
        self : IntervalUnion[T, N, int]
            The interval union where only safe intervals at data points where
            the condition is [`True`][True] are preserved.
        """

        if where is True:
            return self

        return self.union(
            Interval.empty(self._lower.dtype, self.n).preserve_only_where(where)
        )

    def contains(
        self, other: np.ndarray[S, np.dtype[T]]
    ) -> np.ndarray[S, np.dtype[np.bool]]:
        """
        Check if this interval union contains the elements of the `other` array.

        Parameters
        ----------
        other : np.ndarray[S, np.dtype[T]]
            The array whose elements' membership in interval union should be
            checked

        Returns
        -------
        contains : np.ndarray[S, np.dtype[np.bool]]
            The pointwise result of the contains check
        """

        other_flat: np.ndarray[tuple[int], np.dtype[np.unsignedinteger]] = (
            to_total_order(other).flatten()
        )

        (u, n) = self._lower.shape
        is_contained = np.zeros((n,), dtype=np.dtype(bool))

        for i in range(u):
            is_contained |= (other_flat >= to_total_order(self._lower[i])) & (
                other_flat <= to_total_order(self._upper[i])
            )

        return _reshape(is_contained, other.shape)

    def _pick_simple(
        self, prediction: np.ndarray[S, np.dtype[T]]
    ) -> np.ndarray[S, np.dtype[T]]:
        # simple pick:
        #  1. if prediction is in the interval, use it
        #  2. otherwise pick the lower bound of the first interval
        pick: np.ndarray[S, np.dtype[T]] = _ensure_array(
            self._lower[0].reshape(prediction.shape), copy=True
        )
        np.copyto(pick, prediction, where=self.contains(prediction), casting="no")
        return pick

    def _pick_random(
        self,
        prediction: np.ndarray[S, np.dtype[T]],
        rng: np.random.Generator,
    ) -> np.ndarray[S, np.dtype[T]]:
        # random pick: pick a random member of the interval union
        if prediction.size == 0:
            return prediction

        (u, n) = self._lower.shape
        interval_width_acc = self.non_empty_width()

        # use endpoint-inclusive sampling since the single-element width is
        #  zero and the full-interval width is maximal (so +1 would overflow)
        pick_indices: np.ndarray[tuple[N], np.dtype[np.unsignedinteger]] = rng.integers(
            interval_width_acc, dtype=interval_width_acc.dtype, endpoint=True
        )

        interval_width_acc[:] = 0

        pick_flat: np.ndarray[tuple[N], np.dtype[T]] = np.empty(
            (n,), dtype=self._lower.dtype
        )

        for i in range(u):
            lt: np.ndarray[tuple[N], np.dtype[np.unsignedinteger]] = to_total_order(
                self._lower[i]
            )
            ut: np.ndarray[tuple[N], np.dtype[np.unsignedinteger]] = to_total_order(
                self._upper[i]
            )
            np.copyto(
                pick_flat,
                from_total_order(
                    lt + (pick_indices - interval_width_acc), dtype=self._lower.dtype
                ),
                where=(
                    (ut >= lt)
                    & (pick_indices >= interval_width_acc)
                    & (pick_indices <= (interval_width_acc + (ut - lt)))
                ),
                casting="no",
            )
            np.add(
                interval_width_acc,
                ut - lt + 1,
                out=interval_width_acc,
                where=(ut >= lt),
            )

        return _reshape(pick_flat, prediction.shape)

    def _pick_more_zeros(
        self, prediction: np.ndarray[S, np.dtype[T]]
    ) -> np.ndarray[S, np.dtype[T]]:
        if prediction.size == 0:
            return prediction

        (_, n) = self._lower.shape

        # (a) if prediction is in the interval, use it
        contains_prediction = self.contains(prediction).flatten()

        # 1. convert everything to bits in total order
        prediction_bits: np.ndarray[
            tuple[Literal[1], int], np.dtype[np.unsignedinteger]
        ] = _reshape(to_total_order(prediction), (1, -1))

        lower: np.ndarray[tuple[U, N], np.dtype[np.unsignedinteger]] = to_total_order(
            self._lower
        )
        upper: np.ndarray[tuple[U, N], np.dtype[np.unsignedinteger]] = to_total_order(
            self._upper
        )
        interval_nonempty = lower <= upper

        # 2. look at the total order binary difference between the interval
        #    bounds and the prediction
        #    we assume that prediction is not inside the interval, since that
        #    is handled separately with the special case (a)
        lower = lower - prediction_bits
        upper = upper - prediction_bits

        # 3. only work with "positive" unsigned values
        negative: np.ndarray[tuple[U, N], np.dtype[np.bool]] = np.less(
            # lower has an unsigned integer dtype, make signed to compare
            lower.astype(lower.dtype.str.replace("u", "i")),
            0,
        )
        np.copyto(lower, ~lower + 1, where=negative, casting="no")
        np.copyto(upper, ~upper + 1, where=negative, casting="no")

        # 4. ensure that lower <= upper also in binary
        flip: np.ndarray[tuple[U, N], np.dtype[np.bool]] = (
            np.greater(lower, upper) & interval_nonempty
        )
        lower, upper = _where(flip, upper, lower), _where(flip, lower, upper)

        # 5. 0b1111...1111
        allbits = np.array(-1, dtype=upper.dtype.str.replace("u", "i")).view(
            upper.dtype
        )

        # 6. if there are several intervals, pick the one with the smallest
        #    lower bound, ensuring that empty intervals are not picked
        least = _where(interval_nonempty, lower, allbits).argmin(axis=0)
        lower, upper = lower[least, np.arange(n)], upper[least, np.arange(n)]
        negative = negative[least, np.arange(n)]
        assert np.all(lower <= upper)

        # 7. count the number of leading zero bits in lower and upper
        lower_lz: np.ndarray[tuple[U, N], np.dtype[np.uint8]] = _count_leading_zeros(
            lower
        )
        upper_lz: np.ndarray[tuple[U, N], np.dtype[np.uint8]] = _count_leading_zeros(
            upper
        )

        # 8. if upper_lz < lower_lz,
        #    i.e. ceil(log2(upper)) > ceil(log2(lower)),
        #    2 ** ceil(log2(lower)) is a tighter upper bound
        #
        #    upper: 0b00..01xxxxxxxxxxx
        #    lower: 0b00..00..01yyyyyyy
        # -> upper: 0b00..00..100000000
        #
        #    we actually end up choosing, if possible (larger than upper above)
        #    upper: 0b00..0010000000000
        #    since ~half the upper bound works well for symmetric intervals
        np.copyto(
            upper,
            (allbits >> _np_minimum(upper_lz + 2, lower_lz)) + 1,
            where=np.less(upper_lz, lower_lz),
            casting="no",
        )

        # 9. count the number of leading zero bits in (lower ^ upper) to find
        #    the most significant bit where lower and upper differ.
        #    Since upper > lower, at this bit i, upper[i] = 1 and lower[i] = 0.
        lxu_lz = _count_leading_zeros(lower ^ upper)

        # 10. pick such that the binary difference starts and ends with a
        #     maximally long sequence of zeros
        #
        #     upper: 0b00..01xxx1zzzzzzz
        #     lower: 0b00..01yyy0wwwwwww
        #   -> pick: 0b00..01xxx10000000
        pick_bits = _ensure_array(upper & ~(allbits >> (lxu_lz + 1))).reshape(1, -1)

        # 11. undo the negation step to allow "negative" unsigned values again
        np.copyto(pick_bits, ~pick_bits + 1, where=negative, casting="no")

        # 12. undo the difference with prediction and enforce the special case
        #     from (a) that prediction values inside the interval are kept as-is
        pick_bits += prediction_bits
        np.copyto(pick_bits, prediction_bits, where=contains_prediction, casting="no")

        # 13. convert everything back from total-ordered bits to value space
        pick: np.ndarray[S, np.dtype[T]] = _reshape(
            from_total_order(pick_bits, prediction.dtype), prediction.shape
        )
        assert np.all(self.contains(pick))

        return pick

    def pick(
        self, prediction: np.ndarray[S, np.dtype[T]]
    ) -> np.ndarray[S, np.dtype[T]]:
        """
        Pick a member of the interval union that minimises the cost of correcting the prediction to be a member of the interval union.

        The metric for minimising the correction cost is unspecified and may be
        approximate.

        Parameters
        ----------
        prediction : np.ndarray[S, np.dtype[T]]
            A prediction for a member of the interval union

        Returns
        -------
        member : np.ndarray[S, np.dtype[T]]
            A member of the interval union
        """

        # ensure that the prediction size and dtype match the interval union
        assert prediction.size == self._lower.shape[1]
        assert prediction.dtype == self._lower.dtype

        # ensure that every element has a non-empty interval
        assert np.all(to_total_order(self._lower[0]) <= to_total_order(self._upper[0]))

        return self._pick_more_zeros(prediction)

    @override
    def __repr__(self) -> str:
        return f"IntervalUnion(lower={self._lower!r}, upper={self._upper!r})"


def _count_leading_zeros(
    x: np.ndarray[S, np.dtype[np.unsignedinteger]],
) -> np.ndarray[S, np.dtype[np.uint8]]:
    """
    https://stackoverflow.com/a/79189999
    """

    nbits = np.iinfo(x.dtype).bits

    assert nbits <= 64

    if nbits <= 16:
        # safe cast from integer type to a larger integer type,
        # then lossless truncation of the number of leading zeros
        return np.subtract(  # type: ignore
            nbits, np.frexp(x.astype(np.uint32, casting="safe"))[1]
        ).astype(np.uint8, casting="unsafe")

    if nbits <= 32:
        # safe cast from integer type to a larger integer type,
        # then lossless truncation of the number of leading zeros
        return np.subtract(  # type: ignore
            nbits, np.frexp(x.astype(np.uint64, casting="safe"))[1]
        ).astype(np.uint8, casting="unsafe")

    # nbits <= 64
    # safe cast from integer type to a larger integer type,
    _, high_exp = np.frexp(x.astype(np.uint64, casting="safe") >> 32)
    _, low_exp = np.frexp(x.astype(np.uint64, casting="safe") & 0xFFFFFFFF)
    # then lossless truncation of the number of leading zeros
    high_exp += 32
    exp = _ensure_array(low_exp)
    np.copyto(exp, high_exp, where=(high_exp > 32), casting="no")
    return (nbits - exp).astype(np.uint8, casting="unsafe")



"""
Private compatibility functions, mostly wrappers around numpy, that ensure
equivalent behaviour for all supported dtypes and provide good type hints.
"""

__all__ = [
    "_place",
    "_symmetric_modulo",
    "_minimum_zero_sign_sensitive",
    "_maximum_zero_sign_sensitive",
    "_where",
    "_reshape",
    "_broadcast_to",
    "_stack",
    "_ensure_array",
    "_ones",
    "_zeros",
    "_logical_and",
    "_sliding_window_view",
    "_is_sign_negative_number",
    "_is_negative_zero",
    "_is_sign_positive_number",
    "_is_positive_zero",
    "_is_of_dtype",
    "_is_of_shape",
    "_floating_pi",
    "_floating_e",
]

from collections.abc import Sequence
from typing import Literal, TypeGuard, TypeVar, overload

import numpy as np
from numpy.lib.stride_tricks import sliding_window_view

from ._float128 import _float128_dtype, _float128_e, _float128_pi, _float128_type
from .typing import TB, F, Fi, S, Si, T, Ti

N = TypeVar("N", bound=int, covariant=True)
""" Any [`int`][int] (covariant). """


# wrapper around np.place that also works for numpy_quaddtype
# FIXME: https://github.com/numpy/numpy-user-dtypes/issues/236
def _place(
    a: np.ndarray[S, np.dtype[TB]],
    mask: np.ndarray[S, np.dtype[np.bool]],
    vals: np.ndarray[tuple[int], np.dtype[TB]],
) -> None:
    if (type(a) is not _float128_type) and (
        not isinstance(a, np.ndarray) or a.dtype != _float128_dtype
    ):
        return np.place(a, mask, vals)

    return np.put(a, np.flatnonzero(mask), vals)


# wrapper around np.mod(p, q) that guarantees that the result is in [-q/2, q/2]
@overload
def _symmetric_modulo(
    p: np.ndarray[S, np.dtype[F]], q: np.ndarray[S, np.dtype[F]]
) -> np.ndarray[S, np.dtype[F]]: ...


@overload
def _symmetric_modulo(p: Fi, q: Fi) -> Fi: ...


def _symmetric_modulo(p, q):
    q2 = np.divide(q, 2)
    out = _ensure_array(np.add(p, q2))
    np.mod(out, q, out=out)
    np.subtract(out, q2, out=out)
    return out


# wrapper around np.minimum that also works for +0.0 and -0.0
@overload
def _minimum_zero_sign_sensitive(
    a: Ti, b: np.ndarray[S, np.dtype[Ti]]
) -> np.ndarray[S, np.dtype[Ti]]: ...


@overload
def _minimum_zero_sign_sensitive(
    a: np.ndarray[S, np.dtype[Ti]], b: Ti
) -> np.ndarray[S, np.dtype[Ti]]: ...


@overload
def _minimum_zero_sign_sensitive(
    a: np.ndarray[S, np.dtype[T]], b: np.ndarray[S, np.dtype[T]]
) -> np.ndarray[S, np.dtype[T]]: ...


def _minimum_zero_sign_sensitive(a, b):
    a = _ensure_array(a)
    b = _ensure_array(b)
    minimum = np.minimum(a, b)
    if np.issubdtype(a.dtype, np.integer) and np.issubdtype(b.dtype, np.integer):
        return minimum
    minimum_array = _ensure_array(minimum)
    a = _broadcast_to(
        a.astype(minimum_array.dtype, casting="safe"), minimum_array.shape
    )
    b = _broadcast_to(
        b.astype(minimum_array.dtype, casting="safe"), minimum_array.shape
    )
    np.copyto(
        minimum_array,
        a,
        where=((minimum == 0) & (np.signbit(a) > np.signbit(b))),
        casting="no",
    )
    np.copyto(
        minimum_array,
        b,
        where=((minimum == 0) & (np.signbit(a) < np.signbit(b))),
        casting="no",
    )
    return minimum_array


# wrapper around np.maximum that also works for +0.0 and -0.0
@overload
def _maximum_zero_sign_sensitive(
    a: Ti, b: np.ndarray[S, np.dtype[Ti]]
) -> np.ndarray[S, np.dtype[Ti]]: ...


@overload
def _maximum_zero_sign_sensitive(
    a: np.ndarray[S, np.dtype[Ti]], b: Ti
) -> np.ndarray[S, np.dtype[Ti]]: ...


@overload
def _maximum_zero_sign_sensitive(
    a: np.ndarray[S, np.dtype[T]], b: np.ndarray[S, np.dtype[T]]
) -> np.ndarray[S, np.dtype[T]]: ...


def _maximum_zero_sign_sensitive(a, b):
    a = _ensure_array(a)
    b = _ensure_array(b)
    maximum = np.maximum(a, b)
    if np.issubdtype(a.dtype, np.integer) and np.issubdtype(b.dtype, np.integer):
        return maximum
    maximum_array = _ensure_array(maximum)
    a = _broadcast_to(
        a.astype(maximum_array.dtype, casting="safe"), maximum_array.shape
    )
    b = _broadcast_to(
        b.astype(maximum_array.dtype, casting="safe"), maximum_array.shape
    )
    np.copyto(
        maximum_array,
        a,
        where=((maximum == 0) & (np.signbit(a) < np.signbit(b))),
        casting="no",
    )
    np.copyto(
        maximum_array,
        b,
        where=((maximum == 0) & (np.signbit(a) > np.signbit(b))),
        casting="no",
    )
    return maximum_array


# wrapper around np.where but with better type hints
@overload
def _where(
    cond: np.ndarray[S, np.dtype[np.bool]],
    a: np.ndarray[S, np.dtype[TB]],
    b: np.ndarray[S, np.dtype[TB]],
) -> np.ndarray[S, np.dtype[TB]]: ...


@overload
def _where(cond: bool, a: Ti, b: Ti) -> Ti: ...


def _where(cond, a, b):
    return np.where(cond, a, b)


# wrapper around np.reshape but with better type hints
def _reshape(
    a: np.ndarray[tuple[int, ...], np.dtype[TB]], shape: Si
) -> np.ndarray[Si, np.dtype[TB]]:
    return np.reshape(a, shape)


# wrapper around np.broadcast_to but with better type hints
@overload
def _broadcast_to(
    a: np.ndarray[tuple[int, ...], np.dtype[TB]], shape: Si
) -> np.ndarray[Si, np.dtype[TB]]: ...


@overload
def _broadcast_to(a: Ti, shape: Si) -> np.ndarray[Si, np.dtype[Ti]]: ...


def _broadcast_to(a, shape):
    return np.broadcast_to(a, shape)


# wrapper around np.stack but with better type hints
@overload
def _stack(
    arrays: tuple[np.ndarray[tuple[N], np.dtype[T]]],
) -> np.ndarray[tuple[Literal[2], N], np.dtype[T]]: ...


@overload
def _stack(
    arrays: Sequence[np.ndarray[tuple[N], np.dtype[T]]],
) -> np.ndarray[tuple[int, N], np.dtype[T]]: ...


def _stack(arrays):
    return np.stack(arrays)


# wrapper around np.array(a, copy=(copy=None)) but with better type hints
def _ensure_array(
    a: np.ndarray[S, np.dtype[TB]], copy: None | bool = None
) -> np.ndarray[S, np.dtype[TB]]:
    return np.array(a, copy=copy)


# wrapper around np.ones but with better type hints
def _ones(shape: Si, dtype: np.dtype[TB]) -> np.ndarray[Si, np.dtype[TB]]:
    return np.ones(shape, dtype=dtype)


# wrapper around np.zeros but with better type hints
def _zeros(shape: Si, dtype: np.dtype[TB]) -> np.ndarray[Si, np.dtype[TB]]:
    return np.zeros(shape, dtype=dtype)


# wrapper around np.logical_and with better type hints
def _logical_and(
    a: np.ndarray[S, np.dtype[np.bool]],
    b: Literal[True] | np.ndarray[S, np.dtype[np.bool]],
) -> np.ndarray[S, np.dtype[np.bool]]:
    return a & b  # type: ignore


# wrapper around np.lib.stride_tricks.sliding_window_view with better type hints
def _sliding_window_view(
    a: np.ndarray[tuple[int, ...], np.dtype[TB]],
    window_shape: int | tuple[int, ...],
    *,
    axis: int | tuple[int, ...],
    writeable: Literal[False],
) -> np.ndarray[tuple[int, ...], np.dtype[TB]]:
    return sliding_window_view(
        a,
        window_shape,
        # the docs say that tuple[int, ...] is allowed here
        axis=axis,  # type: ignore
        writeable=writeable,
    )


# wrapper around a < 0 that also works for -0.0 (is negative) but excludes NaNs
def _is_sign_negative_number(
    a: np.ndarray[S, np.dtype[F]],
) -> np.ndarray[S, np.dtype[np.bool]]:
    # check not just for a < 0 but also for a == -0.0
    return np.less_equal(a, 0) & (np.copysign(1, a) == -1)


# check for x == -0.0
def _is_negative_zero(
    a: np.ndarray[S, np.dtype[F]],
) -> np.ndarray[S, np.dtype[np.bool]]:
    return (a == 0) & (np.copysign(1, a) == -1)


# wrapper around a > 0 that also works for -0.0 (is not positive) but excludes
#  NaNs
def _is_sign_positive_number(
    a: np.ndarray[S, np.dtype[F]],
) -> np.ndarray[S, np.dtype[np.bool]]:
    # check not just for a > 0 but also for a == +0.0
    return np.greater_equal(a, 0) & (np.copysign(1, a) == +1)


# check for x == +0.0
def _is_positive_zero(
    a: np.ndarray[S, np.dtype[F]],
) -> np.ndarray[S, np.dtype[np.bool]]:
    return (a == 0) & (np.copysign(1, a) == +1)


# type guard for x.dtype == dtype
def _is_of_dtype(
    x: np.ndarray[S, np.dtype[np.number]], dtype: np.dtype[T]
) -> TypeGuard[np.ndarray[S, np.dtype[T]]]:
    return x.dtype == dtype


# type guard for x.shape == shape
def _is_of_shape(
    x: np.ndarray[tuple[int, ...], np.dtype[T]], shape: Si
) -> TypeGuard[np.ndarray[Si, np.dtype[T]]]:
    return x.shape == shape


# wrapper around np.pi, of the dtype, that also works for numpy_quaddtype
def _floating_pi(dtype: np.dtype[F]) -> F:
    if isinstance(_float128_pi, dtype.type):
        return _float128_pi
    return dtype.type(np.pi)


# wrapper around np.e, of the dtype, that also works for numpy_quaddtype
def _floating_e(dtype: np.dtype[F]) -> F:
    if isinstance(_float128_e, dtype.type):
        return _float128_e
    return dtype.type(np.e)



from collections.abc import Callable, Mapping
from typing import TypeVar

import numpy as np
from typing_extensions import override  # MSPV 3.12

from ....utils._compat import _broadcast_to
from ....utils.bindings import Parameter
from ..typing import F, Fi, Ns, Ps, np_sndarray
from .abc import AnyExpr, EmptyExpr, Expr

Ei = TypeVar("Ei", bound=AnyExpr)
""" Any numpy [`Expr`][...abc.Expr] (invariant). """


class ScalarFoldedConstant(EmptyExpr):
    __slots__: tuple[str, ...] = ("_const",)
    _const: np.number

    def __init__(self, const: np.number) -> None:
        self._const = const[()] if isinstance(const, np.ndarray) else const  # type: ignore

    @staticmethod
    def from_folded(e: Ei | Fi) -> "Ei | ScalarFoldedConstant":
        return e if isinstance(e, Expr) else ScalarFoldedConstant(e)

    @property
    @override
    def args(self) -> tuple[()]:
        return ()

    @override
    def with_args(self) -> "ScalarFoldedConstant":
        return ScalarFoldedConstant(self._const)

    @override
    def constant_fold(self, dtype: np.dtype[F]) -> F | AnyExpr:
        assert isinstance(self._const, dtype.type)
        return self._const

    @override
    def eval(
        self,
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> np.ndarray[tuple[Ps], np.dtype[F]]:
        assert isinstance(self._const, Xs.dtype.type)
        return _broadcast_to(self._const, Xs.shape[:1])

    @override
    def compute_data_bounds_unchecked(
        self,
        expr_lower: np.ndarray[tuple[Ps], np.dtype[F]],
        expr_upper: np.ndarray[tuple[Ps], np.dtype[F]],
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> tuple[
        np_sndarray[Ps, Ns, np.dtype[F]],
        np_sndarray[Ps, Ns, np.dtype[F]],
    ]:
        assert False, "folded constants have no data bounds"

    @staticmethod
    def constant_fold_unary(
        expr: AnyExpr,
        dtype: np.dtype[F],
        m: Callable[[F], F],
        rm: Callable[[AnyExpr], AnyExpr],
    ) -> F | AnyExpr:
        fexpr: F | AnyExpr = expr.constant_fold(dtype)
        if isinstance(fexpr, Expr):
            return rm(fexpr)
        return m(fexpr)

    @staticmethod
    def constant_fold_binary(
        left: AnyExpr,
        right: AnyExpr,
        dtype: np.dtype[F],
        m: Callable[[F, F], F],
        rm: Callable[[AnyExpr, AnyExpr], AnyExpr],
    ) -> F | AnyExpr:
        fleft = left.constant_fold(dtype)
        fright = right.constant_fold(dtype)

        if not (isinstance(fleft, Expr) or isinstance(fright, Expr)):
            return m(fleft, fright)

        fleft = ScalarFoldedConstant.from_folded(fleft)
        fright = ScalarFoldedConstant.from_folded(fright)

        return rm(fleft, fright)

    @staticmethod
    def constant_fold_expr(expr: AnyExpr, dtype: np.dtype[F]) -> AnyExpr:
        return ScalarFoldedConstant.from_folded(expr.constant_fold(dtype))

    @override
    def __repr__(self) -> str:
        return f"({self._const!r})"



"""
Logical any (or) combinator safeguard.
"""

__all__ = ["AnySafeguard"]

from abc import ABC
from collections.abc import Collection, Set
from typing import ClassVar, Literal

import numpy as np
from typing_extensions import override  # MSPV 3.12

from ...utils.bindings import Bindings, Parameter
from ...utils.error import TypeCheckError, ctx
from ...utils.intervals import Interval, IntervalUnion
from ...utils.typing import JSON, S, T
from ..abc import Safeguard
from ..pointwise.abc import PointwiseSafeguard
from ..stencil import BoundaryCondition, NeighbourhoodAxis
from ..stencil.abc import StencilSafeguard


class AnySafeguard(Safeguard):
    """
    The `AnySafeguard` guarantees that, for each element, at least one of the
    combined safeguards' guarantees is upheld.

    At the moment, only pointwise and stencil safeguards and combinations
    thereof can be combined by this any-combinator. The combinator is a
    pointwise or a stencil safeguard, depending on the safeguards it combines.

    Parameters
    ----------
    safeguards : Collection[dict[str, JSON] | PointwiseSafeguard | StencilSafeguard]
        At least one safeguard configuration [`dict`][dict]s or already
        initialized [`PointwiseSafeguard`][....pointwise.abc.PointwiseSafeguard]
        or [`StencilSafeguard`][....stencil.abc.StencilSafeguard].

    Raises
    ------
    TypeCheckError
        if any parameter has the wrong type.
    ValueError
        if the `safeguards` collection is empty.
    ...
        if instantiating a safeguard raises an exception.
    """

    __slots__: tuple[str, ...] = ()

    kind: ClassVar[str] = "any"

    def __init__(
        self,
        *,
        safeguards: Collection[dict[str, JSON] | PointwiseSafeguard | StencilSafeguard],
    ) -> None:
        pass

    def __new__(  # type: ignore
        cls,
        *,
        safeguards: Collection[dict[str, JSON] | PointwiseSafeguard | StencilSafeguard],
    ) -> "_AnyPointwiseSafeguard | _AnyStencilSafeguard":
        from ... import SafeguardKind  # noqa: PLC0415

        with ctx.safeguardty(cls):
            with ctx.parameter("safeguards"):
                TypeCheckError.check_instance_or_raise(safeguards, Collection)

                if len(safeguards) <= 0:
                    raise (
                        ValueError("can only combine over at least one safeguard") | ctx
                    )

                safeguards_: list[PointwiseSafeguard | StencilSafeguard] = []
                safeguard: dict[str, JSON] | Safeguard
                for i, safeguard in enumerate(safeguards):
                    with ctx.index(i):
                        TypeCheckError.check_instance_or_raise(
                            safeguard, dict | PointwiseSafeguard | StencilSafeguard
                        )
                        if isinstance(safeguard, dict):
                            safeguard = SafeguardKind.from_config(safeguard)
                        TypeCheckError.check_instance_or_raise(
                            safeguard, PointwiseSafeguard | StencilSafeguard
                        )
                        safeguards_.append(safeguard)  # type: ignore

        pointwise_safeguards_: list[PointwiseSafeguard] = [
            safeguard
            for safeguard in safeguards_
            if isinstance(safeguard, PointwiseSafeguard)
        ]

        if len(pointwise_safeguards_) == len(safeguards_):
            return _AnyPointwiseSafeguard(*pointwise_safeguards_)
        else:
            return _AnyStencilSafeguard(*safeguards_)

    @property
    def safeguards(self) -> tuple[PointwiseSafeguard | StencilSafeguard, ...]:  # type: ignore
        """
        The set of safeguards that this any combinator has been configured to
        uphold.
        """

        ...

    @property
    @override
    def late_bound(self) -> Set[Parameter]:  # type: ignore
        """
        The set of late-bound parameters that this safeguard has.

        Late-bound parameters are only bound when checking and applying the
        safeguard, in contrast to the normal early-bound parameters that are
        configured during safeguard initialisation.

        Late-bound parameters can be used for parameters that depend on the
        specific data that is to be safeguarded.
        """

        ...

    @override
    def check(  # type: ignore
        self,
        data: np.ndarray[S, np.dtype[T]],
        prediction: np.ndarray[S, np.dtype[T]],
        *,
        late_bound: Bindings,
        where: Literal[True] | np.ndarray[S, np.dtype[np.bool]] = True,
    ) -> bool:
        """
        Check if, for all elements, any of the combined safeguards succeed the
        check.

        Parameters
        ----------
        data : np.ndarray[S, np.dtype[T]]
            Original data array, relative to which the `prediction` is checked.
        prediction : np.ndarray[S, np.dtype[T]]
            Prediction for the `data` array.
        late_bound : Bindings
            Bindings for late-bound parameters, including for this safeguard.
        where : Literal[True] | np.ndarray[S, np.dtype[np.bool]]
            Only check at data points where the condition is [`True`][True].

        Returns
        -------
        ok : bool
            `True` if the check succeeded.

        Raises
        ------
        ...
            if checking a safeguard raises an exception.
        """

        ...

    def check_pointwise(  # type: ignore
        self,
        data: np.ndarray[S, np.dtype[T]],
        prediction: np.ndarray[S, np.dtype[T]],
        *,
        late_bound: Bindings,
        where: Literal[True] | np.ndarray[S, np.dtype[np.bool]] = True,
    ) -> np.ndarray[S, np.dtype[np.bool]]:
        """
        Check for which elements at least one of the combined safeguards
        succeeds the check.

        Parameters
        ----------
        data : np.ndarray[S, np.dtype[T]]
            Original data array, relative to which the `prediction` is checked.
        prediction : np.ndarray[S, np.dtype[T]]
            Prediction for the `data` array.
        late_bound : Bindings
            Bindings for late-bound parameters, including for this safeguard.
        where : Literal[True] | np.ndarray[S, np.dtype[np.bool]]
            Only check at data points where the condition is [`True`][True].

        Returns
        -------
        ok : np.ndarray[S, np.dtype[np.bool]]
            Pointwise, `True` if the check succeeded for this element.

        Raises
        ------
        ...
            if checking a safeguard raises an exception.
        """

        ...

    def compute_safe_intervals(  # type: ignore
        self,
        data: np.ndarray[S, np.dtype[T]],
        *,
        late_bound: Bindings,
        where: Literal[True] | np.ndarray[S, np.dtype[np.bool]] = True,
    ) -> IntervalUnion[T, int, int]:
        """
        Compute the union of the safe intervals of the combined safeguards,
        i.e. where at least one is safe.

        Parameters
        ----------
        data : np.ndarray[S, np.dtype[T]]
            Data for which the safe intervals should be computed.
        late_bound : Bindings
            Bindings for late-bound parameters, including for this safeguard.
        where : Literal[True] | np.ndarray[S, np.dtype[np.bool]]
            Only compute the safe intervals at pointwise checks where the
            condition is [`True`][True].

        Returns
        -------
        intervals : IntervalUnion[T, int, int]
            Union of safe intervals.

        Raises
        ------
        ...
            if computing the safe intervals for a safeguard raises an exception.
        """

        ...

    def compute_footprint(  # type: ignore
        self,
        foot: np.ndarray[S, np.dtype[np.bool]],
        *,
        late_bound: Bindings,
        where: Literal[True] | np.ndarray[S, np.dtype[np.bool]] = True,
    ) -> np.ndarray[S, np.dtype[np.bool]]:
        """
        Compute the footprint of the `foot` array, e.g. for expanding data
        points into the pointwise checks that they contribute to.

        Parameters
        ----------
        foot : np.ndarray[S, np.dtype[np.bool]]
            Array for which the footprint is computed.
        late_bound : Bindings
            Bindings for late-bound parameters, including for this safeguard.
        where : Literal[True] | np.ndarray[S, np.dtype[np.bool]]
            Only compute the inverse footprint at pointwise checks where the
            condition is [`True`][True].

        Returns
        -------
        print : np.ndarray[S, np.dtype[np.bool]]
            The footprint of the `foot` array.
        """

        ...

    def compute_inverse_footprint(  # type: ignore
        self,
        foot: np.ndarray[S, np.dtype[np.bool]],
        *,
        late_bound: Bindings,
        where: Literal[True] | np.ndarray[S, np.dtype[np.bool]] = True,
    ) -> np.ndarray[S, np.dtype[np.bool]]:
        """
        Compute the inverse footprint of the `foot` array, e.g. for expanding
        pointwise check fails into the points that could have contributed to
        the failures.

        Parameters
        ----------
        foot : np.ndarray[S, np.dtype[np.bool]]
            Array for which the inverse footprint is computed.
        late_bound : Bindings
            Bindings for late-bound parameters, including for this safeguard.
        where : Literal[True] | np.ndarray[S, np.dtype[np.bool]]
            Only compute the inverse footprint at pointwise checks where the
            condition is [`True`][True].

        Returns
        -------
        print : np.ndarray[S, np.dtype[np.bool]]
            The inverse footprint of the `foot` array.
        """

        ...

    @override
    def get_config(self) -> dict[str, JSON]:  # type: ignore
        """
        Returns the configuration of the safeguard.

        Returns
        -------
        config : dict
            Configuration of the safeguard.
        """

        ...


class _AnySafeguardBase(ABC):
    __slots__: tuple[str, ...] = ()

    kind: ClassVar[str] = "any"

    @property
    def safeguards(self) -> tuple[PointwiseSafeguard | StencilSafeguard, ...]:
        return self._safeguards  # type: ignore

    @property
    def late_bound(self) -> Set[Parameter]:
        return frozenset(b for s in self.safeguards for b in s.late_bound)

    def check_pointwise(
        self,
        data: np.ndarray[S, np.dtype[T]],
        prediction: np.ndarray[S, np.dtype[T]],
        *,
        late_bound: Bindings,
        where: Literal[True] | np.ndarray[S, np.dtype[np.bool]] = True,
    ) -> np.ndarray[S, np.dtype[np.bool]]:
        front, *tail = self.safeguards

        # pointwise check succeeds if, per point, at least one safeguards'
        #  pointwise check succeeds
        ok = front.check_pointwise(data, prediction, late_bound=late_bound, where=where)

        for safeguard in tail:
            ok |= safeguard.check_pointwise(
                data, prediction, late_bound=late_bound, where=where
            )

        return ok

    def compute_footprint(
        self,
        foot: np.ndarray[S, np.dtype[np.bool]],
        *,
        late_bound: Bindings,
        where: Literal[True] | np.ndarray[S, np.dtype[np.bool]] = True,
    ) -> np.ndarray[S, np.dtype[np.bool]]:
        front, *tail = self.safeguards

        # not all safeguards (just one per point) need to contribute to the
        #  footprint
        # however, we cannot be more clever here without knowing the data and
        #  going through the entire safe interval computation, which would
        #  defeat the entire point of this cheaper method
        # so, conservatively, all safeguards contribute to the footprint and
        #  the union of their footprints forms the combined footprint
        footprint = front.compute_footprint(foot, late_bound=late_bound, where=where)

        for safeguard in tail:
            footprint |= safeguard.compute_footprint(
                foot, late_bound=late_bound, where=where
            )

        return footprint

    def compute_inverse_footprint(
        self,
        foot: np.ndarray[S, np.dtype[np.bool]],
        *,
        late_bound: Bindings,
        where: Literal[True] | np.ndarray[S, np.dtype[np.bool]] = True,
    ) -> np.ndarray[S, np.dtype[np.bool]]:
        front, *tail = self.safeguards

        # not all safeguards (just one per point) need to contribute to the
        #  footprint
        # however, we cannot be more clever here without knowing the data and
        #  going through the entire safe interval computation, which would
        #  defeat the entire point of this cheaper method
        # so, conservatively, all safeguards contribute to the footprint and
        #  the union of their footprints forms the combined footprint
        footprint = front.compute_inverse_footprint(
            foot, late_bound=late_bound, where=where
        )

        for safeguard in tail:
            footprint |= safeguard.compute_inverse_footprint(
                foot, late_bound=late_bound, where=where
            )

        return footprint

    def get_config(self) -> dict[str, JSON]:
        return dict(
            kind=type(self).kind,
            safeguards=[safeguard.get_config() for safeguard in self.safeguards],
        )

    @override
    def __repr__(self) -> str:
        return f"{AnySafeguard.__name__}(safeguards={list(self.safeguards)!r})"


class _AnyPointwiseSafeguard(_AnySafeguardBase, PointwiseSafeguard):
    __slots__: tuple[str, ...] = ("_safeguards",)
    _safeguards: tuple[PointwiseSafeguard, ...]

    def __init__(self, *safeguards: PointwiseSafeguard) -> None:
        for safeguard in safeguards:
            assert isinstance(safeguard, PointwiseSafeguard), (
                f"{safeguard!r} is not a pointwise safeguard"
            )
        self._safeguards = safeguards

    @override
    def compute_safe_intervals(
        self,
        data: np.ndarray[S, np.dtype[T]],
        *,
        late_bound: Bindings,
        where: Literal[True] | np.ndarray[S, np.dtype[np.bool]] = True,
    ) -> IntervalUnion[T, int, int]:
        front, *tail = self.safeguards

        # for only-pointwise safeguards, the union over any safeguard's safe
        #  intervals is safe
        valid = front.compute_safe_intervals(data, late_bound=late_bound, where=where)

        for safeguard in tail:
            valid = valid.union(
                safeguard.compute_safe_intervals(
                    data, late_bound=late_bound, where=where
                )
            )

        return valid


class _AnyStencilSafeguard(_AnySafeguardBase, StencilSafeguard):
    __slots__: tuple[str, ...] = ("_safeguards",)
    _safeguards: tuple[PointwiseSafeguard | StencilSafeguard, ...]

    def __init__(self, *safeguards: PointwiseSafeguard | StencilSafeguard) -> None:
        for safeguard in safeguards:
            assert isinstance(safeguard, PointwiseSafeguard | StencilSafeguard), (
                f"{safeguard!r} is not a pointwise or stencil safeguard"
            )
        self._safeguards = safeguards

    @override
    def compute_check_neighbourhood_for_data_shape(
        self, data_shape: tuple[int, ...]
    ) -> tuple[dict[BoundaryCondition, NeighbourhoodAxis], ...]:
        neighbourhood: list[dict[BoundaryCondition, NeighbourhoodAxis]] = [
            dict() for _ in data_shape
        ]

        for safeguard in self._safeguards:
            if not isinstance(safeguard, StencilSafeguard):
                continue

            safeguard_neighbourhood = (
                safeguard.compute_check_neighbourhood_for_data_shape(data_shape)
            )

            for i, sn in enumerate(safeguard_neighbourhood):
                ni = neighbourhood[i]

                for b, s in sn.items():
                    if b in ni:
                        neighbourhood[i][b] = NeighbourhoodAxis(
                            max(ni[b].before, s.before), max(ni[b].after, s.after)
                        )
                    else:
                        neighbourhood[i][b] = s

        return tuple(neighbourhood)

    @override
    def compute_safe_intervals(
        self,
        data: np.ndarray[S, np.dtype[T]],
        *,
        late_bound: Bindings,
        where: Literal[True] | np.ndarray[S, np.dtype[np.bool]] = True,
    ) -> IntervalUnion[T, int, int]:
        pointwise: list[PointwiseSafeguard] = []
        stencil: list[StencilSafeguard] = []
        for safeguard in self.safeguards:
            assert isinstance(safeguard, PointwiseSafeguard | StencilSafeguard)
            if isinstance(safeguard, PointwiseSafeguard):
                pointwise.append(safeguard)
            else:
                stencil.append(safeguard)

        valids = [
            safeguard.compute_safe_intervals(data, late_bound=late_bound, where=where)
            for safeguard in stencil
        ]

        # we cannot pointwise pick safe intervals from different stencil
        #  safeguards, since the stencil requirements may be violated by
        #  valid neighbouring picks
        # very conservatively, the any combinator could be implemented using
        #  the all combinator
        # less conservatively, we try to pick one safeguard per point that we
        #  want to satisfy and then intersect all of their non-pointwise
        #  requirements
        # as an optimisation, we first combine all pointwise safeguards to
        #  favour picking their union of safe intervals

        has_pointwise = len(pointwise) > 0
        if has_pointwise:
            front, *tail = pointwise

            valid_pointwise = front.compute_safe_intervals(
                data, late_bound=late_bound, where=where
            )

            for safeguard in tail:
                valid_pointwise = valid_pointwise.union(
                    safeguard.compute_safe_intervals(
                        data, late_bound=late_bound, where=where
                    )
                )

            # we move the pointwise interval first to prioritise it later
            valids.insert(0, valid_pointwise)

        # simple heuristic: pick the safeguard with the largest interval
        selector = np.argmax([v.non_empty_width() for v in valids], axis=0).reshape(
            data.shape
        )

        valid: IntervalUnion[T, int, int] = Interval.full_like(data).into_union()

        # compute the safe intervals only where needed and then intersect
        # the `where` condition ensures that we only compute restrictions for
        #  the data points for which a safeguard was chosen (and that other
        #  points do not impose such requirements)
        # but that the computed requirements can extend beyond the pointwise
        for i, safeguard in enumerate(stencil):
            where_i = (selector == (i + has_pointwise)) & where
            if np.any(where_i):
                valid = valid.intersect(
                    safeguard.compute_safe_intervals(
                        data, late_bound=late_bound, where=where_i
                    )
                )

        # finally, also include the combined pointwise safe intervals, if
        #  selected
        if has_pointwise:
            where_i = (selector == 0) & where
            valid = valid.intersect(
                valid_pointwise.preserve_only_where(where_i.flatten())
            )

        return valid



from collections.abc import Callable
from enum import Enum, auto
from warnings import warn

import numpy as np

from ...utils._compat import (
    _broadcast_to,
    _ensure_array,
    _is_sign_negative_number,
    _is_sign_positive_number,
)
from ...utils.error import QuantityOfInterestRuntimeWarning
from .typing import Ci, F, J, Ns, Ps, Ps2, np_sndarray, np_sndarray2


def guarantee_arg_within_expr_bounds(
    expr: Callable[
        [np.ndarray[tuple[Ps], np.dtype[F]]], np.ndarray[tuple[Ps], np.dtype[F]]
    ],
    exprv: np.ndarray[tuple[Ps], np.dtype[F]],
    argv: np.ndarray[tuple[Ps], np.dtype[F]],
    argv_bound_guess: np.ndarray[tuple[Ps], np.dtype[F]],
    expr_lower: np.ndarray[tuple[Ps], np.dtype[F]],
    expr_upper: np.ndarray[tuple[Ps], np.dtype[F]],
) -> np.ndarray[tuple[Ps], np.dtype[F]]:
    """
    Ensure that `argv_bound_guess`, a guess for a lower or upper bound on
    argument `argv`, meets the lower and upper bounds `expr_lower` and
    `expr_upper` on the expression `expr`, where `exprv = expr(argv)`.

    Parameters
    ----------
    expr : Callable[[np.ndarray[tuple[Ps], np.dtype[F]]], np.ndarray[tuple[Ps], np.dtype[F]]]
        Evaluate the expression, given the argument `argv`.
    exprv : np.ndarray[tuple[Ps], np.dtype[F]]
        Evaluation of the expression on the argument `argv`.
    argv : np.ndarray[tuple[Ps], np.dtype[F]]
        Pointwise expression argument.
    argv_bound_guess : np.ndarray[tuple[Ps], np.dtype[F]]
        Provided guess for the bound on the argument `argv`.
    expr_lower : np.ndarray[tuple[Ps], np.dtype[F]]
        Pointwise lower bound on the expression, must be less than or equal to
        `exprv`.
    expr_upper : np.ndarray[tuple[Ps], np.dtype[F]]
        Pointwise upper bound on the expression, must be greater than or equal
        to `exprv`.

    Returns
    -------
    argv_bound_guess : np.ndarray[Ps, np.dtype[F]]
        Refined bound that guarantees that
        `expr_lower <= expr(argv_bound_guess) <= expr_upper` or
        `isnan(exprv) & isnan(expr(argv_bound_guess))`.
    """

    return _guarantee_data_within_expr_bounds_inner(
        expr,
        exprv,
        argv,
        argv_bound_guess,
        expr_lower,
        expr_upper,
        warn_on_bounds_exceeded=False,
    )


def guarantee_stacked_arg_within_expr_bounds(
    expr: Callable[
        [np.ndarray[tuple[J, Ps], np.dtype[F]]], np.ndarray[tuple[J, Ps], np.dtype[F]]
    ],
    exprv: np.ndarray[tuple[J, Ps], np.dtype[F]],
    argv: np.ndarray[tuple[J, Ps], np.dtype[F]],
    argv_bound_guess: np.ndarray[tuple[J, Ps], np.dtype[F]],
    expr_lower: np.ndarray[tuple[J, Ps], np.dtype[F]],
    expr_upper: np.ndarray[tuple[J, Ps], np.dtype[F]],
) -> np.ndarray[tuple[J, Ps], np.dtype[F]]:
    """
    Ensure that `argv_bound_guess`, a guess for a lower or upper bound on
    argument `argv`, meets the lower and upper bounds `expr_lower` and
    `expr_upper` on the expression `expr`, where `exprv = expr(argv)`.

    Parameters
    ----------
    expr : Callable[[np.ndarray[tuple[J, Ps], np.dtype[F]]], np.ndarray[tuple[J, Ps], np.dtype[F]]]
        Evaluate the expression, given the argument `argv`.
    exprv : np.ndarray[tuple[J, Ps], np.dtype[F]]
        Evaluation of the expression on the argument `argv`.
    argv : np.ndarray[tuple[J, Ps], np.dtype[F]]
        Pointwise expression argument.
    argv_bound_guess : np.ndarray[tuple[J, Ps], np.dtype[F]]
        Provided guess for the bound on the argument `argv`.
    expr_lower : np.ndarray[tuple[J, Ps], np.dtype[F]]
        Pointwise lower bound on the expression, must be less than or equal to
        `exprv`.
    expr_upper : np.ndarray[tuple[J, Ps], np.dtype[F]]
        Pointwise upper bound on the expression, must be greater than or equal
        to `exprv`.

    Returns
    -------
    argv_bound_guess : np.ndarray[tuple[J, Ps], np.dtype[F]]
        Refined bound that guarantees that
        `expr_lower <= expr(argv_bound_guess) <= expr_upper` or
        `isnan(exprv) & isnan(expr(argv_bound_guess))`.
    """

    return _guarantee_data_within_expr_bounds_inner(
        expr,
        exprv,
        argv,
        argv_bound_guess,
        expr_lower,
        expr_upper,
        warn_on_bounds_exceeded=False,
    )


def guarantee_data_within_expr_bounds(
    expr: Callable[
        [np_sndarray[Ps, Ns, np.dtype[F]]],
        np.ndarray[tuple[Ps], np.dtype[F]],
    ],
    exprv: np.ndarray[tuple[Ps], np.dtype[F]],
    Xs: np_sndarray[Ps, Ns, np.dtype[F]],
    Xs_bound_guess: np_sndarray[Ps, Ns, np.dtype[F]],
    expr_lower: np.ndarray[tuple[Ps], np.dtype[F]],
    expr_upper: np.ndarray[tuple[Ps], np.dtype[F]],
    *,
    warn_on_bounds_exceeded: bool,
) -> np_sndarray[Ps, Ns, np.dtype[F]]:
    """
    Ensure that `Xs_bound_guess`, a guess for a lower or upper bound on `Xs`,
    meets the lower and upper bounds `expr_lower` and `expr_upper` on the
    expression `expr`, where `exprv = expr(Xs)`.

    Parameters
    ----------
    expr : Callable[[np_sndarray[Ps, Ns, np.dtype[F]]], np.ndarray[tuple[Ps], np.dtype[F]]]
        Evaluate the expression, given the stencil-extended data `Xs`.
    exprv : np.ndarray[tuple[Ps], np.dtype[F]]
        Evaluation of the expression on the stencil-extended data `Xs`.
    Xs : np_sndarray[Ps, Ns, np.dtype[F]]
        Stencil-extended data.
    Xs_bound_guess : np_sndarray[Ps, Ns, np.dtype[F]]
        Provided guess for the bound on the stencil-extended data `Xs`.
    expr_lower : np.ndarray[tuple[Ps], np.dtype[F]]
        Pointwise lower bound on the expression, must be less than or equal to
        `exprv`.
    expr_upper : np.ndarray[tuple[Ps], np.dtype[F]]
        Pointwise upper bound on the expression, must be greater than or equal
        to `exprv`.
    warn_on_bounds_exceeded : bool
        If [`True`][True], a warning is emitted when the `Xs_bound_guess` does
        not already meet the `expr_lower` and `expr_upper` bounds.

    Returns
    -------
    Xs_bound_guess : np_sndarray[Ps, Ns, np.dtype[F]]
        Refined bound that guarantees that
        `expr_lower <= expr(Xs_bound_guess) <= expr_upper` or
        `isnan(exprv) & isnan(expr(Xs_bound_guess))`.
    """

    return _guarantee_data_within_expr_bounds_inner(
        expr,
        exprv,
        Xs,
        Xs_bound_guess,
        expr_lower,
        expr_upper,
        warn_on_bounds_exceeded=warn_on_bounds_exceeded,
    )


def _guarantee_data_within_expr_bounds_inner(
    expr: Callable[
        [np_sndarray2[Ps2, Ns, np.dtype[F]]],
        np.ndarray[Ps2, np.dtype[F]],
    ],
    exprv: np.ndarray[Ps2, np.dtype[F]],
    Xs: np_sndarray2[Ps2, Ns, np.dtype[F]],
    Xs_bound_guess: np_sndarray2[Ps2, Ns, np.dtype[F]],
    expr_lower: np.ndarray[Ps2, np.dtype[F]],
    expr_upper: np.ndarray[Ps2, np.dtype[F]],
    *,
    warn_on_bounds_exceeded: bool,
) -> np_sndarray2[Ps2, Ns, np.dtype[F]]:
    exprv = _ensure_array(exprv)
    Xs = _ensure_array(Xs)
    Xs_bound_guess = _ensure_array(Xs_bound_guess, copy=True)
    expr_lower = _ensure_array(expr_lower)
    expr_upper = _ensure_array(expr_upper)

    assert exprv.dtype == Xs.dtype
    assert Xs_bound_guess.dtype == Xs.dtype
    assert expr_lower.dtype == Xs.dtype
    assert expr_upper.dtype == Xs.dtype

    # check if any derived expression exceeds the expression bounds
    def exceeds_expr_bounds(
        Xs_bound_guess: np_sndarray2[Ps2, Ns, np.dtype[F]],
    ) -> np_sndarray2[Ps2, Ns, np.dtype[np.bool]]:
        exprv_Xs_bound_guess = expr(Xs_bound_guess)

        in_bounds: np.ndarray[Ps2, np.dtype[np.bool]] = (
            np.isnan(exprv) & np.isnan(exprv_Xs_bound_guess)
        ) | (
            np.greater_equal(exprv_Xs_bound_guess, expr_lower)
            # if lower >= +0.0, then guess must be >= +0.0
            & (
                ~_is_sign_positive_number(expr_lower)
                | _is_sign_positive_number(exprv_Xs_bound_guess)
            )
            & np.less_equal(exprv_Xs_bound_guess, expr_upper)
            # if upper <= -0.0, then guess must be <= -0.0
            & (
                ~_is_sign_negative_number(expr_upper)
                | _is_sign_negative_number(exprv_Xs_bound_guess)
            )
        )

        bounds_exceeded: np_sndarray2[Ps2, Ns, np.dtype[np.bool]] = _broadcast_to(
            (~in_bounds).reshape(exprv.shape + (1,) * (Xs.ndim - exprv.ndim)),
            Xs.shape,
        )
        return bounds_exceeded

    for _ in range(3):
        bounds_exceeded = exceeds_expr_bounds(Xs_bound_guess)

        if not np.any(bounds_exceeded):
            return Xs_bound_guess

        if warn_on_bounds_exceeded:
            warn_on_bounds_exceeded = False
            warn(
                "guaranteed data bounds do not meet the expression bounds",
                category=QuantityOfInterestRuntimeWarning,
            )

        # nudge the guess towards the data by 1 ULP
        np.nextafter(Xs_bound_guess, Xs, out=Xs_bound_guess, where=bounds_exceeded)

    Xs_diff = _ensure_array(np.nan_to_num(Xs_bound_guess - Xs))

    # exponential backoff for the distance
    backoff = Xs.dtype.type(0.5)

    for _ in range(6):
        bounds_exceeded = exceeds_expr_bounds(Xs_bound_guess)

        if not np.any(bounds_exceeded):
            return Xs_bound_guess

        # shove the guess towards the data by exponentially reducing the
        #  difference
        Xs_diff *= backoff
        backoff = np.divide(backoff, 2)
        np.add(Xs, Xs_diff, out=Xs_bound_guess, where=bounds_exceeded)

    warn(
        "data bounds required excessive nudging",
        category=QuantityOfInterestRuntimeWarning,
    )

    return _ensure_array(Xs, copy=True)


class DataBounds(Enum):
    # not yet checked, will likely need to be corrected
    unchecked = auto()
    # already checked, check should succeed
    checked = auto()
    # it is impossible for anything to go wrong, only use for wrappers that forward
    infallible = auto()


def checked_data_bounds(func: Ci) -> Ci:
    setattr(func, "_checked_data_bounds", DataBounds.checked)
    return func


def data_bounds(bounds: DataBounds) -> Callable[[Ci], Ci]:
    def data_bounds(func: Ci) -> Ci:
        setattr(func, "_checked_data_bounds", bounds)
        return func

    return data_bounds


def data_bounds_checks(func: Callable) -> DataBounds:
    return getattr(func, "_checked_data_bounds", DataBounds.unchecked)



"""
Error bounds that can be guaranteed by various safeguards.
"""

__all__ = ["ErrorBound"]

from enum import Enum, auto
from typing import Never, assert_never

import numpy as np

from ..utils._compat import _ensure_array, _where
from ..utils.cast import from_float, from_total_order, to_float, to_total_order
from ..utils.error import ctx
from ..utils.typing import F, S, T


class ErrorBound(Enum):
    """
    Different types of error bounds that can be guaranteed by various
    safeguards, including:

    - [`ErrorBoundSafeguard`][...pointwise.eb.ErrorBoundSafeguard]
    - [`PointwiseQuantityOfInterestErrorBoundSafeguard`][...pointwise.qoi.eb.PointwiseQuantityOfInterestErrorBoundSafeguard]
    - [`StencilQuantityOfInterestErrorBoundSafeguard`][...stencil.qoi.eb.StencilQuantityOfInterestErrorBoundSafeguard]
    """

    abs = auto()
    r"""
    Absolute error bound, which guarantees that the pointwise absolute error is
    less than or equal to the provided bound $\epsilon_{abs}$:

    \[
    |x - \hat{x}| \leq \epsilon_{abs}
    \]

    or equivalently

    \[
    (x - \epsilon_{abs}) \leq \hat{x} \leq (x + \epsilon_{abs})
    \]

    for a finite $\epsilon_{abs} \geq 0$.
    """

    rel = auto()
    r"""
    Relative error bound, which guarantees that the pointwise relative error is
    less than or equal to the provided bound $\epsilon_{rel}$:

    \[
    |x - \hat{x}| \leq |x| \cdot \epsilon_{rel}
    \]

    or equivalently

    \[
    (x - |x| \cdot \epsilon_{rel}) \leq \hat{x} \leq (x + |x| \cdot \epsilon_{rel})
    \]

    for a finite $\epsilon_{rel} \geq 0$.

    The relative error bound preserves zero values with the same bit pattern.
    """

    ratio = auto()
    r"""
    Ratio error bound, which guarantees that the ratios between the original
    and the corrected values as well as their inverse ratios are less than or
    equal to the provided bound $\epsilon_{ratio}$:

    \[
        \left\{\begin{array}{lr}
            0 \quad &\text{if } x = \hat{x} = 0 \\
            \inf \quad &\text{if } \text{sign}(x) \neq \text{sign}(\hat{x}) \\
            |\log(|x|) - \log(|\hat{x}|)| \quad &\text{otherwise}
        \end{array}\right\} \leq \log(\epsilon_{ratio})
    \]

    or equivalently

    \[
    \begin{split}
        (x \mathbin{/} \epsilon_{ratio}) \leq \hat{x} \leq (x \cdot \epsilon_{ratio}) \quad &\text{if } x \geq 0 \\
        (x \cdot \epsilon_{ratio}) \leq \hat{x} \leq (x \mathbin{/} \epsilon_{ratio}) \quad &\text{otherwise}
    \end{split}
    \]

    for a finite $\epsilon_{ratio} \geq 1$.

    Since the $\epsilon_{ratio}$ bound is finite, ratio error bound also
    guarantees that the sign of each corrected value matches the sign of each
    original value and that a corrected value is zero if and only if it is zero
    in the original data.

    The ratio error bound is sometimes also known as a decimal error bound[^1]
    [^2] if the ratio is expressed as the difference in orders of magnitude. A
    decimal error bound of e.g. $2$ (two orders of magnitude difference / x100
    ratio) can be expressed using
    $\epsilon_{ratio} = {10}^{\epsilon_{decimal}}$.

    The ratio error bound can also be used to guarantee a relative-like error
    bound, e.g. $\epsilon_{ratio} = 1.02$ corresponds to a $2\%$ relative-like
    error bound.

    [^1]: Gustafson, J. L., & Yonemoto, I. T. (2017). Beating floating-point at
        its Own Game: Posit Arithmetic. *Supercomputing Frontiers and
        Innovations*, 4(2). Available from:
        [doi:10.14529/jsfi170206](https://doi.org/10.14529/jsfi170206).

    [^2]: Klöwer, M., Düben, P. D., & Palmer, T. N. (2019). Posits as an
        alternative to floats for weather and climate models. *CoNGA'19:
        Proceedings of the Conference for Next Generation Arithmetic 2019*, 1-8.
        Available from:
        [doi:10.1145/3316279.3316281](https://doi.org/10.1145/3316279.3316281).
    """


def _check_error_bound(
    type: ErrorBound,
    eb: int | float | np.ndarray[tuple[()] | S, np.dtype[F]],
) -> None | Never:
    """
    Check if the error bound value `eb` is valid for the error bound `type`.

    Parameters
    ----------
    type : ErrorBound
        The error bound type.
    eb : int | float | np.ndarray[tuple[()] | S, np.dtype[F]]
        The error bound value.

    Raises
    ------
    ValueError
        if the error bound `value` is invalid.
    """

    match type:
        case ErrorBound.abs:
            if not np.all(eb >= 0):
                raise (
                    ValueError("must be non-negative for an absolute error bound") | ctx
                )
        case ErrorBound.rel:
            if not np.all(eb >= 0):
                raise (
                    ValueError("must be non-negative for a relative error bound") | ctx
                )
        case ErrorBound.ratio:
            if not np.all(eb >= 1):
                raise ValueError("must be >= 1 for a ratio error bound") | ctx
        case _:
            assert_never(type)

    if (not isinstance(eb, int)) and (not np.all(np.isfinite(eb))):
        raise ValueError("must be finite") | ctx

    return None


def _compute_finite_absolute_error_bound(
    type: ErrorBound,
    eb: np.ndarray[tuple[()] | S, np.dtype[F]],
    data_float: np.ndarray[S, np.dtype[F]],
) -> np.ndarray[tuple[()] | S, np.dtype[F]]:
    """
    Compute the actual absolute error bound value, against which the error is
    compared against, from the error bound `eb` and the data.

    The computation is only defined for finite data elements, i.e. the produced
    error bound should not be used for non-finite data elements.

    Parameters
    ----------
    type : ErrorBound
        The error bound type.
    eb : np.ndarray[tuple[()] | S, np.dtype[F]]
        The error bound value. It must have already been checked using
        `_check_error_bound`.
    data_float : np.ndarray[S, np.dtype[F]]
        The original data array in floating-point representation.

    Returns
    -------
    eb_abs : np.ndarray[tuple[()] | S, np.dtype[F]]
        The absolute error bound value.

        The error bound is scalar if `eb` is scalar and the error bound does
        not depend on the data, otherwise it has the same shape as the data.
    """

    match type:
        case ErrorBound.abs:
            return eb
        case ErrorBound.rel:
            with np.errstate(
                divide="ignore", over="ignore", under="ignore", invalid="ignore"
            ):
                eb_rel_as_abs = np.nan_to_num(np.abs(data_float) * eb)
            assert np.all((eb_rel_as_abs >= 0) & np.isfinite(eb_rel_as_abs))
            return eb_rel_as_abs
        case ErrorBound.ratio:
            return eb
        case _:
            assert_never(type)


@np.errstate(divide="ignore", over="ignore", under="ignore", invalid="ignore")
def _compute_finite_absolute_error(
    type: ErrorBound,
    data_float: np.ndarray[S, np.dtype[F]],
    prediction_float: np.ndarray[S, np.dtype[F]],
) -> np.ndarray[S, np.dtype[F]]:
    """
    Compute the absolute error value, which is compared against the absolute
    error bound, from the data and prediction arrays.

    The computation is only defined for finite data elements, i.e. the produced
    error should not be used for non-finite data or prediction elements.

    Parameters
    ----------
    type : ErrorBound
        The error bound type.
    data_float : np.ndarray[S, np.dtype[F]]
        The original data array in floating-point representation.
    prediction_float : np.ndarray[S, np.dtype[F]]
        The prediction data array in floating-point representation.

    Returns
    -------
    err_abs : np.ndarray[S, np.dtype[F]]
        The absolute error value between the original and prediction data.
    """

    match type:
        # for the relative error bound, we also just return the absolute error
        #  here since the error bound will include the data value that together
        #  form a relative error bound
        case ErrorBound.abs | ErrorBound.rel:
            return np.abs(np.subtract(data_float, prediction_float))
        case ErrorBound.ratio:
            err_abs: np.ndarray[S, np.dtype[F]] = _ensure_array(
                np.divide(prediction_float, data_float)
            )
            np.divide(
                data_float,
                prediction_float,
                out=err_abs,
                where=(np.abs(data_float) > np.abs(prediction_float)),
            )
            err_abs[(data_float == 0) & (prediction_float == 0)] = 0
            err_abs[np.sign(data_float) != np.sign(prediction_float)] = np.inf
            return err_abs
        case _:
            assert_never(type)


def _apply_finite_error_bound(
    type: ErrorBound,
    eb: np.ndarray[tuple[()] | S, np.dtype[F]],
    data: np.ndarray[S, np.dtype[T]],
    data_float: np.ndarray[S, np.dtype[F]],
) -> tuple[np.ndarray[S, np.dtype[T]], np.ndarray[S, np.dtype[T]]]:
    """
    Apply the error bound `eb` to the `data` to produce the inclusive lower and
    upper bounds within which the error bound is satisfied.

    This function calls `_compute_finite_absolute_error_bound` internally.

    The computation is only defined for finite data elements, i.e. the produced
    bounds should not be used for non-finite data elements.

    Parameters
    ----------
    type : ErrorBound
        The error bound type.
    eb : np.ndarray[tuple[()] | S, np.dtype[F]]
        The error bound value. It must have already been checked using
        `_check_error_bound`.
    data : np.ndarray[S, np.dtype[T]]
        The original data array.
    data_float : np.ndarray[S, np.dtype[F]]
        The original data array in floating-point representation.

    Returns
    -------
    lower, upper : tuple[np.ndarray[S, np.dtype[T]], np.ndarray[S, np.dtype[T]]]
        The lower and upper bounds within which the error bound is satisfied.
    """

    eb_abs = _compute_finite_absolute_error_bound(type, eb, data_float)

    match type:
        case ErrorBound.abs | ErrorBound.rel:
            with np.errstate(
                divide="ignore", over="ignore", under="ignore", invalid="ignore"
            ):
                lower: np.ndarray[S, np.dtype[T]] = from_float(
                    data_float - eb_abs, data.dtype
                )
                upper: np.ndarray[S, np.dtype[T]] = from_float(
                    data_float + eb_abs, data.dtype
                )

            # correct rounding errors in the lower and upper bound
            with np.errstate(
                divide="ignore", over="ignore", under="ignore", invalid="ignore"
            ):
                lower_outside_eb: np.ndarray[S, np.dtype[np.bool]] = np.greater(
                    data_float - to_float(lower, ftype=data_float.dtype), eb_abs
                )
                upper_outside_eb: np.ndarray[S, np.dtype[np.bool]] = np.greater(
                    to_float(upper, ftype=data_float.dtype) - data_float, eb_abs
                )

            lower = _ensure_array(
                from_total_order(
                    np.add(to_total_order(lower), lower_outside_eb),  # type: ignore
                    data.dtype,
                )
            )
            upper = _ensure_array(
                from_total_order(
                    np.subtract(to_total_order(upper), upper_outside_eb),  # type: ignore
                    data.dtype,
                )
            )

            # a zero-error bound must preserve exactly, e.g. even for -0.0
            if np.any(eb == 0):
                np.copyto(lower, data, where=(eb == 0), casting="no")
                np.copyto(upper, data, where=(eb == 0), casting="no")
        case ErrorBound.ratio:
            with np.errstate(
                divide="ignore", over="ignore", under="ignore", invalid="ignore"
            ):
                data_mul, data_div = (
                    from_float(data_float * eb_abs, data.dtype),
                    from_float(data_float / eb_abs, data.dtype),
                )
            lower = _where(np.less(data, 0), data_mul, data_div)
            upper = _where(np.less(data, 0), data_div, data_mul)

            # correct rounding errors in the lower and upper bound
            with np.errstate(
                divide="ignore", over="ignore", under="ignore", invalid="ignore"
            ):
                lower_outside_eb = (
                    np.abs(
                        _where(
                            np.less(data, 0),
                            to_float(lower, ftype=data_float.dtype) / data_float,
                            data_float / to_float(lower, ftype=data_float.dtype),
                        )
                    )
                    > eb_abs
                )
                upper_outside_eb = (
                    np.abs(
                        _where(
                            np.less(data, 0),
                            data_float / to_float(upper, ftype=data_float.dtype),
                            to_float(upper, ftype=data_float.dtype) / data_float,
                        )
                    )
                    > eb_abs
                )

            lower = _ensure_array(
                from_total_order(
                    np.add(to_total_order(lower), lower_outside_eb),  # type: ignore
                    data.dtype,
                )
            )
            upper = _ensure_array(
                from_total_order(
                    np.subtract(to_total_order(upper), upper_outside_eb),  # type: ignore
                    data.dtype,
                )
            )

            # a ratio of 1 bound must preserve exactly, e.g. even for -0.0
            if np.any(eb == 1):
                np.copyto(lower, data, where=(eb == 1), casting="no")
                np.copyto(upper, data, where=(eb == 1), casting="no")
        case _:
            assert_never(type)

    if type in (ErrorBound.rel, ErrorBound.ratio):
        # special case zero to handle +0.0 and -0.0
        np.copyto(lower, data, where=(data == 0), casting="no")
        np.copyto(upper, data, where=(data == 0), casting="no")

    return (lower, upper)


def _apply_finite_qoi_error_bound(
    type: ErrorBound,
    eb: np.ndarray[tuple[()] | S, np.dtype[F]],
    qoi_float: np.ndarray[S, np.dtype[F]],
) -> tuple[np.ndarray[S, np.dtype[F]], np.ndarray[S, np.dtype[F]]]:
    """
    Apply the error bound `eb` to the QoI array to produce the inclusive lower
    and upper bounds within which the error bound is satisfied.

    This function calls `_compute_finite_absolute_error_bound` internally.

    Unlike `_apply_finite_error_bound`, which is applied in the original data
    space, this function is applied in floating-point QoI space and uses a
    different approach for correcting rounding errors with nudging.

    The computation is only defined for finite data elements, i.e. the produced
    bounds should not be used for non-finite QoI elements.

    Parameters
    ----------
    type : ErrorBound
        The error bound type.
    eb : np.ndarray[tuple[()] | S, np.dtype[F]]
        The error bound value. It must have already been checked using
        `_check_error_bound`.
    qoi_float : np.ndarray[S, np.dtype[F]]
        The floating-point QoI array.

    Returns
    -------
    lower, upper : tuple[np.ndarray[S, np.dtype[F]], np.ndarray[S, np.dtype[F]]]
        The lower and upper bounds within which the error bound is satisfied.
    """

    eb_abs = _compute_finite_absolute_error_bound(type, eb, qoi_float)

    match type:
        case ErrorBound.abs | ErrorBound.rel:
            with np.errstate(
                divide="ignore", over="ignore", under="ignore", invalid="ignore"
            ):
                lower: np.ndarray[S, np.dtype[F]] = _ensure_array(
                    np.subtract(qoi_float, eb_abs)
                )
                upper: np.ndarray[S, np.dtype[F]] = _ensure_array(
                    np.add(qoi_float, eb_abs)
                )

            # optimistically allow both -0.0 and +0.0
            lower[lower == 0] = -0.0
            upper[upper == 0] = +0.0

            # correct rounding errors in the lower and upper bound
            with np.errstate(
                divide="ignore", over="ignore", under="ignore", invalid="ignore"
            ):
                lower_outside_eb: np.ndarray[S, np.dtype[np.bool]] = np.greater(
                    qoi_float - lower, eb_abs
                )
                upper_outside_eb: np.ndarray[S, np.dtype[np.bool]] = np.greater(
                    upper - qoi_float, eb_abs
                )

            # we can nudge with nextafter since the QoIs are floating-point
            np.nextafter(
                lower,
                qoi_float,
                out=lower,
                where=(lower_outside_eb & np.isfinite(qoi_float)),
            )
            np.nextafter(
                upper,
                qoi_float,
                out=upper,
                where=(upper_outside_eb & np.isfinite(qoi_float)),
            )

            # a zero-error bound must preserve exactly, e.g. even for -0.0
            if np.any(eb == 0):
                np.copyto(lower, qoi_float, where=(eb == 0), casting="no")
                np.copyto(upper, qoi_float, where=(eb == 0), casting="no")
        case ErrorBound.ratio:
            with np.errstate(
                divide="ignore", over="ignore", under="ignore", invalid="ignore"
            ):
                data_mul, data_div = (
                    qoi_float * eb_abs,
                    qoi_float / eb_abs,
                )
            lower = _ensure_array(_where(np.less(qoi_float, 0), data_mul, data_div))
            upper = _ensure_array(_where(np.less(qoi_float, 0), data_div, data_mul))

            # correct rounding errors in the lower and upper bound
            with np.errstate(
                divide="ignore", over="ignore", under="ignore", invalid="ignore"
            ):
                lower_outside_eb = (
                    np.abs(
                        _where(
                            qoi_float < 0,
                            lower / qoi_float,
                            qoi_float / lower,
                        )
                    )
                    > eb_abs
                )
                upper_outside_eb = (
                    np.abs(
                        _where(
                            qoi_float < 0,
                            qoi_float / upper,
                            upper / qoi_float,
                        )
                    )
                    > eb_abs
                )

            # we can nudge with nextafter since the QoIs are floating-point
            np.nextafter(
                lower,
                qoi_float,
                out=lower,
                where=(lower_outside_eb & np.isfinite(qoi_float)),
            )
            np.nextafter(
                upper,
                qoi_float,
                out=upper,
                where=(upper_outside_eb & np.isfinite(qoi_float)),
            )

            # a ratio of 1 bound must preserve exactly, e.g. even for -0.0
            if np.any(eb == 1):
                np.copyto(lower, qoi_float, where=(eb == 1), casting="no")
                np.copyto(upper, qoi_float, where=(eb == 1), casting="no")
        case _:
            assert_never(type)

    if type in (ErrorBound.rel, ErrorBound.ratio):
        # special case zero to handle +0.0 and -0.0
        np.copyto(lower, qoi_float, where=(qoi_float == 0), casting="no")
        np.copyto(upper, qoi_float, where=(qoi_float == 0), casting="no")

    return (lower, upper)
